from django.contrib import admin
from django.urls import path
from django.utils.translation import gettext_lazy

from kmuhelper.modules.emails import views
from kmuhelper.modules.emails.models import (
    EMail,
    EMailAttachment,
    Attachment,
    EMailTemplate,
)
from kmuhelper.overrides import CustomModelAdmin, CustomTabularInline

_ = gettext_lazy

#######


@admin.register(Attachment)
class AttachmentAdmin(CustomModelAdmin):
    list_display = [
        "pkfill",
        "filename",
        "description",
        "time_created",
        "autocreated",
        "id",
    ]

    search_fields = (
        "filename",
        "description",
    )

    ordering = ["-time_created"]

    save_on_top = True

    HIDDEN = True

    def get_readonly_fields(self, request, obj=None):
        if obj:
            return ["time_created", "file", "autocreated"]

        return ["time_created", "autocreated"]

    def get_fieldsets(self, request, obj=None):
        default = [
            (_("Datei"), {"fields": ["filename", "file"]}),
            (_("Infos"), {"fields": ["description"]}),
        ]

        if obj:
            return default + [
                (_("Daten"), {"fields": ["autocreated", "time_created"]}),
            ]

        return default

    # Views

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name

        urls = super().get_urls()

        my_urls = [
            path(
                "<path:object_id>/download/",
                self.admin_site.admin_view(views.attachment_download),
                name="%s_%s_download" % info,
            ),
        ]
        return my_urls + urls


class EMailAdminAttachmentInline(CustomTabularInline):
    model = EMailAttachment
    verbose_name = _("Anhang")
    verbose_name_plural = _("Anhänge")
    extra = 0

    show_change_link = True

    autocomplete_fields = ("attachment",)

    # Permissions

    NO_CHANGE = True


@admin.register(EMail)
class EMailAdmin(CustomModelAdmin):
    def get_fieldsets(self, request, obj=None):
        default = [
            (_("Infos"), {"fields": ["subject", "to"]}),
            (
                _("Zusätzliche Empfänger"),
                {"fields": ["cc", "bcc"], "classes": ["collapse"]},
            ),
            (_("Inhalt"), {"fields": ["html_template", "text", "html_context"]}),
        ]

        if obj:
            return default + [
                (_("Zeiten"), {"fields": ["time_created", "time_sent"]}),
            ]

        return default

    readonly_fields = ("time_created", "time_sent")

    ordering = ("sent", "-time_sent", "-time_created")

    list_display = (
        "pkfill",
        "subject",
        "to",
        "html_template",
        "time_created",
        "sent",
        "time_sent",
    )
    list_filter = ("html_template", "sent")

    search_fields = ["subject", "to", "cc", "bcc", "html_context", "text", "notes"]

    inlines = (EMailAdminAttachmentInline,)

    save_on_top = True

    HIDDEN = True

    # Views

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name

        urls = super().get_urls()

        my_urls = [
            path(
                "<path:object_id>/preview/",
                self.admin_site.admin_view(views.email_preview),
                name="%s_%s_preview" % info,
            ),
            path(
                "<path:object_id>/send/",
                self.admin_site.admin_view(views.email_send),
                name="%s_%s_send" % info,
            ),
            path(
                "<path:object_id>/resend/",
                self.admin_site.admin_view(views.email_resend),
                name="%s_%s_resend" % info,
            ),
        ]
        return my_urls + urls

    # Custom save

    def save_model(self, request, obj, form, change):
        if obj:
            obj.is_valid(request)
        super().save_model(request, obj, form, change)


@admin.register(EMailTemplate)
class EMailTemplateAdmin(CustomModelAdmin):
    list_display = ["pkfill", "title", "description", "get_use_link"]

    search_fields = (
        "title",
        "description",
    )

    ordering = ["title"]

    fieldsets = [
        (_("Infos"), {"fields": ["title", "description"]}),
        (_("Inhalt"), {"fields": ["mail_to", "mail_subject", "mail_text"]}),
        (_("Optionen"), {"fields": ["mail_template", "mail_context"]}),
    ]

    save_on_top = True

    HIDDEN = True

    # Views

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name

        urls = super().get_urls()

        my_urls = [
            path(
                "savevars",
                self.admin_site.admin_view(views.emailtemplate_savevars),
                name="%s_%s_savevars" % info,
            ),
            path(
                "resetvars",
                self.admin_site.admin_view(views.emailtemplate_resetvars),
                name="%s_%s_resetvars" % info,
            ),
            path(
                "<path:object_id>/use/",
                self.admin_site.admin_view(views.emailtemplate_use),
                name="%s_%s_use" % info,
            ),
        ]
        return my_urls + urls


#


modeladmins = [
    (EMail, EMailAdmin),
    (Attachment, AttachmentAdmin),
    (EMailTemplate, EMailTemplateAdmin),
]



"""
 Exports a model to a CSV file
"""

import csv

from django.apps import apps
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = "Output the specified model as CSV"

    def add_arguments(self, parser):
        parser.add_argument(
            "model",
            nargs=1,
            type=str,
            help='Model name to export, like <app.model> or "members.Member"',
        )
        parser.add_argument(
            "outfile",
            nargs=1,
            type=str,
            help='Save path, like </path/to/outfile.csv> or "/data/members.csv"',
        )

    def handle(self, *args, **options):
        app_name, model_name = options["model"][0].split(".")
        model = apps.get_model(app_name, model_name)
        field_names = [f.name for f in model._meta.fields]
        writer = csv.writer(
            open(options["outfile"][0], "w"), quoting=csv.QUOTE_ALL, delimiter=","
        )
        writer.writerow(field_names)
        for instance in model.objects.all():
            writer.writerow([str(getattr(instance, f)) for f in field_names])



from django.contrib.admin.models import LogEntry
from django.db import models
from django.utils.translation import gettext_lazy

_ = gettext_lazy


class AdminLogEntryManager(models.Manager):
    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .select_related("content_type", "user")
            .filter(content_type__app_label="kmuhelper")
        )


class AdminLogEntry(LogEntry):
    objects = AdminLogEntryManager()

    admin_icon = "fa-solid fa-clock-rotate-left"

    class Meta:
        proxy = True
        verbose_name = _("Admin-Logeintrag")
        verbose_name_plural = _("Admin-Logeinträge")



# Generated by Django 5.0.7 on 2024-07-17 22:13

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("kmuhelper", "0115_migrate_payment_receiver_address_format"),
    ]

    operations = [
        migrations.AddIndex(
            model_name="customer",
            index=models.Index(
                fields=["woocommerceid"], name="kmuhelper_c_woocomm_b647fd_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="order",
            index=models.Index(
                fields=["woocommerceid"], name="kmuhelper_o_woocomm_8db6e3_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="product",
            index=models.Index(
                fields=["woocommerceid"], name="kmuhelper_p_woocomm_34f8a4_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="productcategory",
            index=models.Index(
                fields=["woocommerceid"], name="kmuhelper_p_woocomm_a29e62_idx"
            ),
        ),
    ]



import uuid

from django.db import models
from django.contrib import admin

from django.utils.translation import gettext_lazy, gettext

from kmuhelper import settings
from kmuhelper.overrides import CustomModel

_ = gettext_lazy


class ApiKey(CustomModel):
    """Model representing an api key"""

    key = models.UUIDField(
        verbose_name=_("Key"),
        default=uuid.uuid4,
        unique=True,
    )
    name = models.CharField(
        verbose_name=_("Name"),
        max_length=100,
        default="",
        blank=True,
    )
    user = models.ForeignKey(
        to=settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
    )

    read = models.BooleanField(
        verbose_name=_("Read permission?"),
        default=True,
    )
    write = models.BooleanField(
        verbose_name=_("Write permission?"),
        default=False,
    )

    objects = models.Manager()

    @admin.display(description=_("Api key"))
    def __str__(self):
        perms = (
            _("read")
            if self.read and not self.write
            else _("write")
            if self.write and not self.read
            else _("read/write")
            if self.read and self.write
            else _("UNUSABLE")
        )
        return f"{self.name} ({perms}; {self.user.username})"

    @admin.display(description=_("Key preview"))
    def key_preview(self):
        """Get the first and last letters of the key"""
        return str(self.key)[:4] + "..." + str(self.key)[-4:]

    def has_perm(self, *args, **kwargs):
        """Shortcut for self.user.has_perm"""
        return self.user.has_perm(*args, **kwargs)

    def has_perms(self, *args, **kwargs):
        """Shortcut for self.user.has_perms"""
        return self.user.has_perms(*args, **kwargs)

    class Meta:
        verbose_name = _("Api key")
        verbose_name_plural = _("Api keys")



# Generated by Django 5.0 on 2024-01-14 16:59
from datetime import datetime

from django.db import migrations, models
from django.utils import timezone

VAT_RATE_REVERSE_MIGRATION = {2.6: 2.5, 3.8: 3.7, 8.1: 7.7}


def copy_values(apps, schema_editor):
    db_alias = schema_editor.connection.alias
    OrderItem = apps.get_model("kmuhelper", "OrderItem")

    for item in (
        OrderItem.objects.using(db_alias).all().select_related("order", "product")
    ):
        item.article_number = item.product.article_number
        item.quantity_description = item.product.quantity_description
        item.name = item.product.name
        item.vat_rate = item.product.vat_rate

        if (
            item.order.date
            < timezone.make_aware(
                datetime(year=2024, month=1, day=1, hour=0, minute=0, second=0)
            )
            and item.vat_rate in VAT_RATE_REVERSE_MIGRATION
        ):
            item.vat_rate = VAT_RATE_REVERSE_MIGRATION[item.vat_rate]

        item.save()


class Migration(migrations.Migration):
    dependencies = [
        ("kmuhelper", "0109_contactperson_is_default_paymentreceiver_is_default"),
    ]

    operations = [
        migrations.AddField(
            model_name="orderitem",
            name="article_number",
            field=models.CharField(
                default="", max_length=25, verbose_name="Artikelnummer"
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="orderitem",
            name="name",
            field=models.CharField(
                default="",
                help_text="Unterstützt i18n: '[:de]Deutsch[:fr]Français[:it]Italiano[:en]English[:]'",
                max_length=500,
                verbose_name="Name",
            ),
        ),
        migrations.AddField(
            model_name="orderitem",
            name="quantity_description",
            field=models.CharField(
                blank=True,
                default="Stück",
                help_text="Unterstützt i18n: '[:de]Deutsch[:fr]Français[:it]Italiano[:en]English[:]'",
                max_length=100,
                verbose_name="Mengenbezeichnung",
            ),
        ),
        migrations.AddField(
            model_name="orderitem",
            name="vat_rate",
            field=models.FloatField(
                choices=[
                    (0.0, "0.0% (Mehrwertsteuerfrei)"),
                    (2.6, "2.6% (Reduzierter Satz)"),
                    (3.8, "3.8% (Sondersatz für Beherbergung)"),
                    (8.1, "8.1% (Normalsatz)"),
                    (2.5, "2.5% (Bis 2023: Reduzierter Satz)"),
                    (3.7, "3.7% (Bis 2023: Sondersatz für Beherbergung)"),
                    (7.7, "7.7% (Bis 2023: Normalsatz)"),
                ],
                default=8.1,
                verbose_name="MwSt-Satz",
            ),
        ),
        migrations.RunPython(
            # This migration cannot be reverted!
            copy_values,
            reverse_code=None,
            elidable=True,
        ),
        migrations.RenameField(
            model_name="orderitem", old_name="product", new_name="linked_product"
        ),
        migrations.AlterField(
            model_name="orderitem",
            name="linked_product",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=models.deletion.SET_NULL,
                to="kmuhelper.product",
                verbose_name="Verknüpftes Produkt",
            ),
        ),
    ]



"Models for KMUHelper settings"

from django import forms
from django.contrib import admin
from django.db import models
from django.templatetags.static import static
from django.utils.html import mark_safe, urlize
from django.utils.translation import gettext_lazy
from kmuhelper.modules.settings.constants import SETTINGS
from kmuhelper.overrides import CustomModel

_ = gettext_lazy

SETTINGTYPES = [
    ("char", _("Text")),
    ("text", _("Mehrzeiliger Text")),
    ("bool", _("Wahrheitswert")),
    ("int", _("Zahl")),
    ("float", _("Fliesskommazahl")),
    ("url", _("Url")),
    ("email", _("E-Mail")),
    ("json", _("JSON Daten")),
]


class SettingBase(CustomModel):
    """Base model for 'Setting' and 'SettingHidden'"""

    id = models.CharField(
        verbose_name=_("ID"),
        max_length=50,
        primary_key=True,
    )
    typ = models.CharField(
        verbose_name=_("Typ"),
        max_length=5,
        default="char",
        choices=SETTINGTYPES,
    )

    content_char = models.CharField(
        verbose_name=_("Inhalt (Text)"),
        max_length=250,
        default="",
        blank=True,
    )
    content_text = models.TextField(
        verbose_name=_("Inhalt (Mehrzeiliger Text)"),
        default="",
        blank=True,
    )
    content_bool = models.BooleanField(
        verbose_name=_("Inhalt (Wahrheitswert)"),
        default=False,
        blank=True,
    )
    content_int = models.IntegerField(
        verbose_name=_("Inhalt (Zahl)"),
        default=0,
        blank=True,
    )
    content_float = models.FloatField(
        verbose_name=_("Inhalt (Fliesskommazahl)"),
        default=0.0,
        blank=True,
    )
    content_url = models.URLField(
        verbose_name=_("Inhalt (Url)"),
        default="",
        blank=True,
    )
    content_email = models.EmailField(
        verbose_name=_("Inhalt (E-Mail)"),
        default="",
        blank=True,
    )

    content_json = models.JSONField(
        verbose_name=_("Inhalt (JSON)"),
        default=dict,
        blank=True,
        null=True,
    )

    @property
    @admin.display(description=_("Inhalt"))
    def content_display(self):
        if self.typ == "bool":
            return mark_safe(
                '<img src="'
                + static(f"admin/img/icon-{'yes' if self.content_bool else 'no'}.svg")
                + '" />'
            )
        if self.typ == "url":
            return mark_safe(urlize(self.content_url))
        return self.content

    @property
    def content(self):
        return getattr(
            self,
            f"content_{self.typ}",
            _("ERROR! Unbekannter Einstellungstyp! Bitte kontaktiere den Entwickler!"),
        )

    @content.setter
    def content(self, var):
        if hasattr(self, f"content_{self.typ}"):
            return setattr(self, f"content_{self.typ}", var)

    class Meta:
        abstract = True

    objects = models.Manager()


class Setting(SettingBase):
    """Model representing an editable setting"""

    @property
    def info(self):
        "Get the info dict for this setting"
        return SETTINGS.get(self.id, {})

    @property
    @admin.display(description=_("Einstellung"))
    def name(self):
        s = self.info.get("name", _("(unbekannte Einstellung)"))
        return mark_safe(s)

    @property
    @admin.display(description=_("Beschreibung"))
    def description(self):
        s = self.info.get("description", _("(unbekannte Einstellung)"))
        s = s.replace("\n", "<br />")
        return mark_safe(s)

    def get_field(self):
        "Get the corresponding form field for this setting"

        opt = {
            "label": self.name,
            "required": False,
            "help_text": self.description,
            "initial": self.content,
        }

        if self.typ == "char":
            return forms.CharField(**opt)
        if self.typ == "text":
            return forms.CharField(widget=forms.Textarea, **opt)
        if self.typ == "bool":
            return forms.BooleanField(**opt)
        if self.typ == "int":
            return forms.IntegerField(**opt)
        if self.typ == "float":
            return forms.FloatField(**opt)
        if self.typ == "url":
            return forms.URLField(**opt)
        if self.typ == "email":
            return forms.EmailField(**opt)
        if self.typ == "json":
            return forms.JSONField(**opt)
        raise ValueError("Unknown setting type! Can't get field!")

    @admin.display(description=_("Einstellung"))
    def __str__(self):
        return str(self.name)

    class Meta:
        verbose_name = _("Einstellung")
        verbose_name_plural = _("Einstellungen")
        default_permissions = ("change",)

    objects = models.Manager()

    ADMIN_ICON = "fa-solid fa-cog"


class SettingHidden(SettingBase):
    """Model representing a hidden setting

    Hidden settings should only be edited through code and are not
    meant to be seen by the user.

    Example usage: WooCommerce authentication data"""

    class Meta:
        verbose_name = _("Versteckte Einstellung")
        verbose_name_plural = _("Versteckte Einstellungen")
        default_permissions = ()



import datetime
import json

from django.contrib.auth.decorators import login_required
from django.core.serializers.json import DjangoJSONEncoder
from django.shortcuts import render
from django.urls import reverse_lazy
from django.utils import timezone
from django.utils.translation import gettext_lazy
from kmuhelper.decorators import require_all_kmuhelper_perms
from kmuhelper.modules.main.models import OrderItem, Order
from kmuhelper.translations import langselect
from kmuhelper.utils import render_error

_ = gettext_lazy

#####


@login_required(login_url=reverse_lazy("kmuhelper:login"))
def stats(request):
    return render(request, "kmuhelper/stats/index.html", {"has_permission": True})


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["view_product", "view_order"])
def stats_products_price(request):
    if not Order.objects.exists():
        return render_error(
            request, status=400, message=_("Keine Bestellungen vorhanden.")
        )

    try:
        from_date = timezone.make_aware(
            datetime.datetime.strptime(str(request.GET.get("from")), "%Y-%m-%d")
        )
    except (ValueError, IndexError):
        from_date = Order.objects.order_by("date").first().date

    try:
        to_date = timezone.make_aware(
            datetime.datetime.strptime(str(request.GET.get("to")), "%Y-%m-%d")
        )
    except (ValueError, IndexError):
        to_date = Order.objects.order_by("date").last().date

    products_sold = {}

    for bp in (
        OrderItem.objects.filter(order__date__gte=from_date, order__date__lte=to_date)
        .order_by("order__date")
        .values("order__date", "quantity")
    ):
        d = datetime.date(
            year=bp["order__date"].year, month=bp["order__date"].month, day=1
        ).isoformat()
        if d in products_sold:
            products_sold[d] += bp["quantity"]
        else:
            products_sold[d] = bp["quantity"]

    products_sold = [{"date": date, "y": products_sold[date]} for date in products_sold]
    products_sold_data = json.dumps(products_sold, cls=DjangoJSONEncoder)

    money_income = {}

    for b in (
        Order.objects.filter(date__gte=from_date, date__lte=to_date)
        .order_by("date")
        .values("date", "cached_sum")
    ):
        d = datetime.date(year=b["date"].year, month=b["date"].month, day=1).isoformat()
        if d in money_income:
            money_income[d] += b["cached_sum"]
        else:
            money_income[d] = b["cached_sum"]

    money_income = [{"date": date, "y": money_income[date]} for date in money_income]
    money_income_data = json.dumps(money_income, cls=DjangoJSONEncoder)

    context = {
        "has_permission": True,
        "products_sold": products_sold_data,
        "money_income": money_income_data,
    }
    return render(request, "kmuhelper/stats/products_price.html", context)


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["view_product", "view_order"])
def best_products(request):
    if not Order.objects.exists():
        return render_error(
            request, status=400, message=_("Keine Bestellungen vorhanden.")
        )

    try:
        if "max" in request.GET:
            max_count = int(request.GET.get("max"))
        else:
            max_count = 20
    except (ValueError, IndexError):
        max_count = 20

    try:
        from_date = timezone.make_aware(
            datetime.datetime.strptime(str(request.GET.get("from")), "%Y-%m-%d")
        )
    except (ValueError, IndexError):
        from_date = Order.objects.order_by("date").first().date

    try:
        to_date = timezone.make_aware(
            datetime.datetime.strptime(str(request.GET.get("to")), "%Y-%m-%d")
        )
    except (ValueError, IndexError):
        to_date = Order.objects.order_by("date").last().date

    products = {}

    for bp in (
        OrderItem.objects.filter(order__date__gte=from_date, order__date__lte=to_date)
        .order_by("linked_product__name")
        .values("linked_product__name", "name", "quantity")
    ):
        product_name = langselect(bp["linked_product__name"] or bp["name"])
        if product_name in products:
            products[product_name] += bp["quantity"]
        else:
            products[product_name] = bp["quantity"]

    products = sorted(products.items(), key=lambda x: x[1], reverse=True)
    products = products[:max_count] if len(products) > max_count else products

    context = {
        "has_permission": True,
        "product_names": [p[0] for p in products],
        "product_counts": [p[1] for p in products],
    }
    return render(request, "kmuhelper/stats/best_products.html", context)



# KMUHelper modules configuration

from django.apps import apps
from django.shortcuts import reverse
from django.utils.translation import gettext_lazy

_ = gettext_lazy

MODULES = {
    "home": {
        "title": _("Home"),
        "viewname": "kmuhelper:home",
        "model_names": [],
    },
    "main": {
        "title": _("Admin"),
        "viewname": "kmuhelper:main-index",
        "model_names": [
            "ContactPerson",
            "Order",
            "Fee",
            "Customer",
            "Supplier",
            "Supply",
            "Note",
            "Product",
            "ProductCategory",
            "PaymentReceiver",
            "PaymentImport",
        ],
    },
    "emails": {
        "title": _("E-Mails"),
        "viewname": "kmuhelper:email-index",
        "model_names": [
            "EMailTemplate",
            "EMail",
            "Attachment",
        ],
    },
    "app": {
        "title": _("App"),
        "viewname": "kmuhelper:app-index",
        "model_names": [
            "App_ToDo",
            "App_Stock",
            "App_Shipping",
            "App_Arrival",
            "App_IncomingPayments",
        ],
    },
    "settings": {
        "title": _("Einstellungen"),
        "viewname": "kmuhelper:settings",
        "model_names": [
            "Setting",
        ],
    },
    "stats": {
        "title": _("Statistiken"),
        "viewname": "kmuhelper:stats",
        "model_names": [],
    },
    "log": {
        "title": _("Log"),
        "viewname": "kmuhelper:log",
        "model_names": [
            "AdminLogEntry",
        ],
    },
}

MODULE_ASSOCIATIONS = {}


# Utils


def get_model(model_name):
    return apps.get_model("kmuhelper", model_name)


def get_models(model_names):
    models = [get_model(model_name) for model_name in model_names]
    return models


def __load():
    for module_name, module in MODULES.items():
        for model in get_models(module.get("model_names")):
            MODULE_ASSOCIATIONS[model._meta.model_name] = module_name


def get_model_from_context(context, model_name=None):
    if model_name is None:
        model_opts = context.get("opts", None)
        if model_opts is None:
            return None
        model_name = model_opts.model_name
    return get_model(model_name)


def get_module_from_context(context, module_name=None):
    if module_name is None:
        model_opts = context.get("opts", None)
        if model_opts is None:
            module_name = "main"
        else:
            model_name = model_opts.model_name

            if MODULE_ASSOCIATIONS == {}:
                __load()

            module_name = MODULE_ASSOCIATIONS.get(model_name, "main")
    return MODULES.get(module_name)


def get_module_home_context(request, module_name):
    module = MODULES.get(module_name)
    return {
        "app_label": "kmuhelper",
        "app_list": [
            {
                "app_label": "kmuhelper",
                "app_url": reverse(module.get("viewname")),
                "has_module_perms": True,
                "models": [
                    {
                        "add_url": reverse(
                            f"admin:kmuhelper_{model._meta.model_name}_add"
                        ),
                        "admin_url": reverse(
                            f"admin:kmuhelper_{model._meta.model_name}_changelist"
                        ),
                        "name": model._meta.verbose_name_plural,
                        "object_name": model._meta.model_name,
                        "perms": {
                            "add": request.user.has_perm(
                                f"kmuhelper.add_{model._meta.model_name}"
                            ),
                            "change": request.user.has_perm(
                                f"kmuhelper.change_{model._meta.model_name}"
                            ),
                            "delete": request.user.has_perm(
                                f"kmuhelper.delete_{model._meta.model_name}"
                            ),
                            "view": request.user.has_perm(
                                f"kmuhelper.view_{model._meta.model_name}"
                            ),
                        },
                        "view_only": not request.user.has_perm(
                            f"kmuhelper.change_{model._meta.model_name}"
                        ),
                    }
                    for model in get_models(module.get("model_names"))
                    if (
                        request.user.has_perm(
                            f"kmuhelper.view_{model._meta.model_name}"
                        )
                        or request.user.has_perm(
                            f"kmuhelper.change_{model._meta.model_name}"
                        )
                    )
                ],
                "name": module.get("title"),
            }
        ],
        "has_permission": True,
        "is_nav_sidebar_enabled": False,
        "is_popup": False,
        "title": module.get("full_title"),
    }


def user_has_module_permission(user, module_name):
    module = MODULES.get(module_name)
    for model_name in module.get("model_names"):
        name = model_name.lower()
        if user.has_perm(f"kmuhelper.view_{name}") or user.has_perm(
            f"kmuhelper.change_{name}"
        ):
            return True
    return False



"""Useful constants used in multiple places"""

from django.utils.translation import gettext_lazy as _

URL_MANUAL = "https://rafaelurben.github.io/django-kmuhelper/manual/"
URL_FAQ = "https://rafaelurben.github.io/django-kmuhelper/manual/faq"

ORDERSTATUS = [
    ("pending", _("Zahlung ausstehend")),
    ("processing", _("In Bearbeitung")),
    ("on-hold", _("In Wartestellung")),
    ("completed", _("Abgeschlossen")),
    ("cancelled", _("Storniert/Abgebrochen")),
    ("refunded", _("Rückerstattet")),
    ("failed", _("Fehlgeschlagen")),
    ("trash", _("Gelöscht")),
]

VAT_RATES = [
    (0.0, _("0.0% (Mehrwertsteuerfrei)")),
    (2.6, _("2.6% (Reduzierter Satz)")),
    (3.8, _("3.8% (Sondersatz für Beherbergung)")),
    (8.1, _("8.1% (Normalsatz)")),
    (2.5, _("2.5% (Bis 2023: Reduzierter Satz)")),
    (3.7, _("3.7% (Bis 2023: Sondersatz für Beherbergung)")),
    (7.7, _("7.7% (Bis 2023: Normalsatz)")),
]
VAT_RATE_DEFAULT = 8.1

PAYMENTMETHODS = [
    ("bacs", _("Überweisung")),
    ("cheque", _("Scheck")),
    ("cod", _("Rechnung / Nachnahme")),
    ("paypal", _("PayPal")),
]

COUNTRIES = [("CH", _("Schweiz")), ("LI", _("Liechtenstein"))]

LANGUAGES = [
    ("de", _("Deutsch [DE]")),
    ("fr", _("Französisch [FR]")),
    ("it", _("Italienisch [IT]")),
    ("en", _("Englisch [EN]")),
]

INVOICE_DISPLAY_MODES = [
    ("business_orders", _("Geschäftlich (Bestellungen)")),
    ("business_services", _("Geschäftlich (Dienstleistungen)")),
    ("club", _("Verein")),
    ("private", _("Privat")),
]

# Addresses

ADDR_FIELDS_WITHOUT_CONTACT = [
    "first_name",
    "last_name",
    "company",
    "address_1",
    "address_2",
    "postcode",
    "city",
    "state",
    "country",
]

ADDR_FIELDS = ADDR_FIELDS_WITHOUT_CONTACT + [
    "email",
    "phone",
]

ADDR_BILLING_FIELDS_WITHOUT_CONTACT = list(
    map(lambda f: f"addr_billing_{f}", ADDR_FIELDS_WITHOUT_CONTACT)
)

ADDR_BILLING_FIELDS = list(map(lambda f: f"addr_billing_{f}", ADDR_FIELDS))

ADDR_BILLING_FIELDS_CATEGORIZED = [
    ("addr_billing_first_name", "addr_billing_last_name"),
    "addr_billing_company",
    ("addr_billing_address_1", "addr_billing_address_2"),
    ("addr_billing_postcode", "addr_billing_city"),
    ("addr_billing_state", "addr_billing_country"),
    ("addr_billing_email", "addr_billing_phone"),
]

ADDR_SHIPPING_FIELDS_WITHOUT_CONTACT = list(
    map(lambda f: f"addr_shipping_{f}", ADDR_FIELDS_WITHOUT_CONTACT)
)

ADDR_SHIPPING_FIELDS = list(map(lambda f: f"addr_shipping_{f}", ADDR_FIELDS))

ADDR_SHIPPING_FIELDS_CATEGORIZED = [
    ("addr_shipping_first_name", "addr_shipping_last_name"),
    "addr_shipping_company",
    ("addr_shipping_address_1", "addr_shipping_address_2"),
    ("addr_shipping_postcode", "addr_shipping_city"),
    ("addr_shipping_state", "addr_shipping_country"),
    ("addr_shipping_email", "addr_shipping_phone"),
]



"""
Tests if there are changes detected that are not yet reflected in a migration
"""

import unittest

from django.core.management import call_command


class GeneralTest(unittest.TestCase):
    def test_detected_changes(self):
        try:
            call_command("makemigrations", "kmuhelper", "--check", "--dry-run")
        except SystemExit:
            self.fail("There are changes that are not yet reflected in a migration!")



# Generated by Django 4.1.6 on 2023-02-20 18:31

import uuid

import django.core.validators
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models

import kmuhelper.external.multi_email_field.fields
import kmuhelper.modules.emails.models
import kmuhelper.modules.main.models

""" Important note!

This migration is a squashed migration. It technically contains all the migrations from 0001 to 0100.

This file has only partially been generated by squashing the migrations with the `manage.py squashmigrations 100` command.
It has been mainly generated by removing all the old migrations and then running `manage.py makemigrations`.

Reason for this "fake squashing" is that squashmigrations did not work properly and still left some
CreateModel, DeleteModel and RenameField operations in the squashed migration.

I am aware that this is not the proper way to squash migrations, but it is the only way I could get it to work.

- Rafael, 2023-02-20
"""


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Attachment",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("filename", models.CharField(max_length=50, verbose_name="Dateiname")),
                (
                    "file",
                    models.FileField(
                        upload_to=kmuhelper.modules.emails.models.getfilepath,
                        verbose_name="Datei",
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True, default="", verbose_name="Beschreibung"
                    ),
                ),
                (
                    "autocreated",
                    models.BooleanField(
                        default=False, verbose_name="Automatisch generiert"
                    ),
                ),
                (
                    "token",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, verbose_name="Token"
                    ),
                ),
                (
                    "time_created",
                    models.DateTimeField(auto_now_add=True, verbose_name="Erstellt um"),
                ),
            ],
            options={
                "verbose_name": "Anhang",
                "verbose_name_plural": "Anhänge",
                "default_permissions": ("add", "change", "view", "delete", "download"),
            },
        ),
        migrations.CreateModel(
            name="ContactPerson",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Auf Rechnung ersichtlich!",
                        max_length=50,
                        verbose_name="Name",
                    ),
                ),
                (
                    "phone",
                    models.CharField(
                        help_text="Auf Rechnung ersichtlich!",
                        max_length=50,
                        verbose_name="Telefon",
                    ),
                ),
                (
                    "email",
                    models.EmailField(
                        help_text="Auf Rechnung ersichtlich!",
                        max_length=254,
                        verbose_name="E-Mail",
                    ),
                ),
            ],
            options={
                "verbose_name": "Ansprechpartner",
                "verbose_name_plural": "Ansprechpartner",
            },
        ),
        migrations.CreateModel(
            name="Customer",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "woocommerceid",
                    models.IntegerField(default=0, verbose_name="WooCommerce ID"),
                ),
                (
                    "email",
                    models.EmailField(
                        blank=True, max_length=254, verbose_name="E-Mail Adresse"
                    ),
                ),
                (
                    "first_name",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Vorname"
                    ),
                ),
                (
                    "last_name",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Nachname"
                    ),
                ),
                (
                    "company",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Firma"
                    ),
                ),
                (
                    "username",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=50,
                        verbose_name="Benutzername",
                    ),
                ),
                (
                    "avatar_url",
                    models.URLField(
                        blank=True, editable=False, verbose_name="Avatar URL"
                    ),
                ),
                (
                    "language",
                    models.CharField(
                        choices=[
                            ("de", "Deutsch [DE]"),
                            ("fr", "Französisch [FR]"),
                            ("it", "Italienisch [IT]"),
                            ("en", "Englisch [EN]"),
                        ],
                        default="de",
                        max_length=2,
                        verbose_name="Sprache",
                    ),
                ),
                (
                    "addr_billing_first_name",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Vorname"
                    ),
                ),
                (
                    "addr_billing_last_name",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Nachname"
                    ),
                ),
                (
                    "addr_billing_company",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Firma"
                    ),
                ),
                (
                    "addr_billing_address_1",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text='Strasse und Hausnummer oder "Postfach"',
                        max_length=250,
                        verbose_name="Adresszeile 1",
                    ),
                ),
                (
                    "addr_billing_address_2",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Wird in QR-Rechnung NICHT verwendet!",
                        max_length=250,
                        verbose_name="Adresszeile 2",
                    ),
                ),
                (
                    "addr_billing_city",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Ort"
                    ),
                ),
                (
                    "addr_billing_state",
                    models.CharField(
                        blank=True, default="", max_length=50, verbose_name="Kanton"
                    ),
                ),
                (
                    "addr_billing_postcode",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=50,
                        verbose_name="Postleitzahl",
                    ),
                ),
                (
                    "addr_billing_country",
                    models.CharField(
                        choices=[("CH", "Schweiz"), ("LI", "Liechtenstein")],
                        default="CH",
                        max_length=2,
                        verbose_name="Land",
                    ),
                ),
                (
                    "addr_billing_email",
                    models.EmailField(
                        blank=True, max_length=254, verbose_name="E-Mail-Adresse"
                    ),
                ),
                (
                    "addr_billing_phone",
                    models.CharField(
                        blank=True, default="", max_length=50, verbose_name="Telefon"
                    ),
                ),
                (
                    "addr_shipping_first_name",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Vorname"
                    ),
                ),
                (
                    "addr_shipping_last_name",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Nachname"
                    ),
                ),
                (
                    "addr_shipping_company",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Firma"
                    ),
                ),
                (
                    "addr_shipping_address_1",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=250,
                        verbose_name="Adresszeile 1",
                    ),
                ),
                (
                    "addr_shipping_address_2",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=250,
                        verbose_name="Adresszeile 2",
                    ),
                ),
                (
                    "addr_shipping_city",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Ort"
                    ),
                ),
                (
                    "addr_shipping_state",
                    models.CharField(
                        blank=True, default="", max_length=50, verbose_name="Kanton"
                    ),
                ),
                (
                    "addr_shipping_postcode",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=50,
                        verbose_name="Postleitzahl",
                    ),
                ),
                (
                    "addr_shipping_country",
                    models.CharField(
                        choices=[("CH", "Schweiz"), ("LI", "Liechtenstein")],
                        default="CH",
                        max_length=2,
                        verbose_name="Land",
                    ),
                ),
                (
                    "addr_shipping_email",
                    models.EmailField(
                        blank=True, max_length=254, verbose_name="E-Mail-Adresse"
                    ),
                ),
                (
                    "addr_shipping_phone",
                    models.CharField(
                        blank=True, default="", max_length=50, verbose_name="Telefon"
                    ),
                ),
                (
                    "website",
                    models.URLField(blank=True, default="", verbose_name="Webseite"),
                ),
                (
                    "note",
                    models.TextField(blank=True, default="", verbose_name="Bemerkung"),
                ),
                (
                    "combine_with",
                    models.ForeignKey(
                        blank=True,
                        help_text="Dies kann nicht widerrufen werden! Werte im aktuellen Kunden werden bevorzugt.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="kmuhelper.customer",
                        verbose_name="Zusammenfügen mit",
                    ),
                ),
            ],
            options={
                "verbose_name": "Kunde",
                "verbose_name_plural": "Kunden",
            },
        ),
        migrations.CreateModel(
            name="EMail",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "subject",
                    models.CharField(
                        help_text="Wird Standardmässig auch als Untertitel verwendet.",
                        max_length=50,
                        verbose_name="Betreff",
                    ),
                ),
                (
                    "to",
                    kmuhelper.external.multi_email_field.fields.MultiEmailField(
                        default=[],
                        help_text="Direkte Empfänger",
                        verbose_name="Empfänger",
                    ),
                ),
                (
                    "cc",
                    kmuhelper.external.multi_email_field.fields.MultiEmailField(
                        blank=True, default="", help_text="Kopie", verbose_name="CC"
                    ),
                ),
                (
                    "bcc",
                    kmuhelper.external.multi_email_field.fields.MultiEmailField(
                        blank=True,
                        default="",
                        help_text="Blindkopie",
                        verbose_name="BCC",
                    ),
                ),
                (
                    "language",
                    models.CharField(
                        choices=[
                            ("de", "Deutsch [DE]"),
                            ("fr", "Französisch [FR]"),
                            ("it", "Italienisch [IT]"),
                            ("en", "Englisch [EN]"),
                        ],
                        default="de",
                        max_length=2,
                        verbose_name="Sprache",
                    ),
                ),
                (
                    "html_template",
                    models.CharField(
                        default="default.html",
                        help_text="Dateiname der Designvorlage unter 'kmuhelper/emails/'.",
                        max_length=50,
                        verbose_name="Designvorlage",
                    ),
                ),
                (
                    "text",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="Hauptinhalt - wird nicht von allen Designvorlagen verwendet. Links und E-Mail Adressen werden automatisch verlinkt.",
                        verbose_name="Text",
                    ),
                ),
                (
                    "html_context",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        help_text="Daten im JSON-Format, mit welchen die Designvorlage befüllt wird.",
                        null=True,
                        verbose_name="Daten",
                    ),
                ),
                (
                    "token",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, verbose_name="Token"
                    ),
                ),
                (
                    "time_created",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="Datum und Zeit der Erstellung dieser E-Mail.",
                        verbose_name="Erstellt am",
                    ),
                ),
                (
                    "time_sent",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="Datum und Zeit des letzten erfolgreichen Sendeversuches.",
                        null=True,
                        verbose_name="Gesendet um",
                    ),
                ),
                ("sent", models.BooleanField(default=False, verbose_name="Gesendet?")),
                (
                    "notes",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="Diese Notizen haben keine Einwirkung auf die E-Mail selbst.",
                        verbose_name="Notizen",
                    ),
                ),
            ],
            options={
                "verbose_name": "E-Mail",
                "verbose_name_plural": "E-Mails",
                "default_permissions": ("add", "change", "view", "delete", "send"),
            },
        ),
        migrations.CreateModel(
            name="EMailTemplate",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=50, verbose_name="Titel")),
                (
                    "description",
                    models.TextField(
                        blank=True, default="", verbose_name="Beschreibung"
                    ),
                ),
                (
                    "mail_to",
                    models.CharField(
                        blank=True, default="", max_length=50, verbose_name="Empfänger"
                    ),
                ),
                (
                    "mail_subject",
                    models.CharField(max_length=50, verbose_name="Betreff"),
                ),
                ("mail_text", models.TextField(verbose_name="Text")),
                (
                    "mail_template",
                    models.CharField(
                        default="default.html",
                        max_length=50,
                        verbose_name="Designvorlage",
                    ),
                ),
                (
                    "mail_context",
                    models.JSONField(
                        blank=True, default=dict, null=True, verbose_name="Daten"
                    ),
                ),
            ],
            options={
                "verbose_name": "E-Mail-Vorlage",
                "verbose_name_plural": "E-Mail-Vorlagen",
            },
        ),
        migrations.CreateModel(
            name="Fee",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        default="Zusätzliche Kosten",
                        help_text="Unterstützt i18n: '[:de]Deutsch[:fr]Français[:it]Italiano[:en]English[:]'",
                        max_length=500,
                        verbose_name="Bezeichnung",
                    ),
                ),
                (
                    "price",
                    models.FloatField(default=0.0, verbose_name="Preis (exkl. MwSt)"),
                ),
                (
                    "vat_rate",
                    models.FloatField(
                        choices=[
                            (0.0, "0.0% (Mehrwertsteuerfrei)"),
                            (2.5, "2.5% (Bis 2023: Reduzierter Satz)"),
                            (2.6, "2.6% (Ab 2024: Reduzierter Satz)"),
                            (3.7, "3.7% (Bis 2023: Sondersatz für Beherbergung)"),
                            (3.8, "3.8% (Ab 2024: Sondersatz für Beherbergung)"),
                            (7.7, "7.7% (Bis 2023: Normalsatz)"),
                            (8.1, "8.1% (Ab 2024: Normalsatz)"),
                        ],
                        default=7.7,
                        verbose_name="MwSt-Satz",
                    ),
                ),
            ],
            options={
                "verbose_name": "Kosten",
                "verbose_name_plural": "Kosten",
            },
        ),
        migrations.CreateModel(
            name="Order",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "woocommerceid",
                    models.IntegerField(default=0, verbose_name="WooCommerce ID"),
                ),
                (
                    "date",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="Datum"
                    ),
                ),
                (
                    "invoice_date",
                    models.DateField(
                        blank=True,
                        default=None,
                        help_text="Wird auch als Startpunkt für die Zahlungskonditionen verwendet. Wird beim Erstellen eines PDF automatisch mit dem aktuellen Datum befüllt.",
                        null=True,
                        verbose_name="Rechnungsdatum",
                    ),
                ),
                (
                    "payment_conditions",
                    models.CharField(
                        default=kmuhelper.modules.main.models.default_payment_conditions,
                        help_text='Skonto und Zahlungsfrist -> <a target="_blank" href="https://rafaelurben.github.io/django-kmuhelper/manual/faq#wie-funktionieren-zahlungskonditionen">FAQ</a>',
                        max_length=16,
                        validators=[
                            django.core.validators.RegexValidator(
                                "^([0-9]+(\\.[0-9]+)?:[0-9]+;)*0:[0-9]+$",
                                "Bitte benutze folgendes Format: 'p:d;p:d' - p = Skonto in %; d = Tage",
                            )
                        ],
                        verbose_name="Zahlungskonditionen",
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("pending", "Zahlung ausstehend"),
                            ("processing", "In Bearbeitung"),
                            ("on-hold", "In Wartestellung"),
                            ("completed", "Abgeschlossen"),
                            ("cancelled", "Storniert/Abgebrochen"),
                            ("refunded", "Rückerstattet"),
                            ("failed", "Fehlgeschlagen"),
                            ("trash", "Gelöscht"),
                        ],
                        default="processing",
                        max_length=11,
                        verbose_name="Status",
                    ),
                ),
                (
                    "is_shipped",
                    models.BooleanField(
                        default=False,
                        help_text='Mehr Infos -> <a target="_blank" href="https://rafaelurben.github.io/django-kmuhelper/manual/faq#was-passiert-wenn-ich-eine-bestellung-als-bezahltversendet-markiere">FAQ</a>',
                        verbose_name="Versendet?",
                    ),
                ),
                (
                    "shipped_on",
                    models.DateField(
                        blank=True, default=None, null=True, verbose_name="Versendet am"
                    ),
                ),
                (
                    "tracking_number",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Bitte gib hier eine Trackingnummer der Schweizer Post ein. (optional)",
                        max_length=25,
                        validators=[
                            django.core.validators.RegexValidator(
                                "^99\\.[0-9]{2}\\.[0-9]{6}\\.[0-9]{8}$",
                                "Bite benutze folgendes Format: 99.xx.xxxxxx.xxxxxxxx",
                            )
                        ],
                        verbose_name="Trackingnummer",
                    ),
                ),
                (
                    "is_removed_from_stock",
                    models.BooleanField(default=False, verbose_name="Ausgelagert?"),
                ),
                (
                    "payment_method",
                    models.CharField(
                        choices=[
                            ("bacs", "Überweisung"),
                            ("cheque", "Scheck"),
                            ("cod", "Rechnung / Nachnahme"),
                            ("paypal", "PayPal"),
                        ],
                        default="cod",
                        max_length=7,
                        verbose_name="Zahlungsmethode",
                    ),
                ),
                (
                    "is_paid",
                    models.BooleanField(
                        default=False,
                        help_text='Mehr Infos -> <a target="_blank" href="https://rafaelurben.github.io/django-kmuhelper/manual/faq#was-passiert-wenn-ich-eine-bestellung-als-bezahltversendet-markiere">FAQ</a>',
                        verbose_name="Bezahlt?",
                    ),
                ),
                (
                    "paid_on",
                    models.DateField(
                        blank=True, default=None, null=True, verbose_name="Bezahlt am"
                    ),
                ),
                (
                    "customer_note",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="Vom Kunden erfasste Notiz.",
                        verbose_name="Kundennotiz",
                    ),
                ),
                (
                    "order_key",
                    models.CharField(
                        blank=True,
                        default=kmuhelper.modules.main.models.default_order_key,
                        max_length=50,
                        verbose_name="Bestellungs-Schlüssel",
                    ),
                ),
                (
                    "addr_billing_first_name",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Vorname"
                    ),
                ),
                (
                    "addr_billing_last_name",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Nachname"
                    ),
                ),
                (
                    "addr_billing_company",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Firma"
                    ),
                ),
                (
                    "addr_billing_address_1",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text='Strasse und Hausnummer oder "Postfach"',
                        max_length=250,
                        verbose_name="Adresszeile 1",
                    ),
                ),
                (
                    "addr_billing_address_2",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Wird in QR-Rechnung NICHT verwendet!",
                        max_length=250,
                        verbose_name="Adresszeile 2",
                    ),
                ),
                (
                    "addr_billing_city",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Ort"
                    ),
                ),
                (
                    "addr_billing_state",
                    models.CharField(
                        blank=True, default="", max_length=50, verbose_name="Kanton"
                    ),
                ),
                (
                    "addr_billing_postcode",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=50,
                        verbose_name="Postleitzahl",
                    ),
                ),
                (
                    "addr_billing_country",
                    models.CharField(
                        choices=[("CH", "Schweiz"), ("LI", "Liechtenstein")],
                        default="CH",
                        max_length=2,
                        verbose_name="Land",
                    ),
                ),
                (
                    "addr_billing_email",
                    models.EmailField(
                        blank=True, max_length=254, verbose_name="E-Mail-Adresse"
                    ),
                ),
                (
                    "addr_billing_phone",
                    models.CharField(
                        blank=True, default="", max_length=50, verbose_name="Telefon"
                    ),
                ),
                (
                    "addr_shipping_first_name",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Vorname"
                    ),
                ),
                (
                    "addr_shipping_last_name",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Nachname"
                    ),
                ),
                (
                    "addr_shipping_company",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Firma"
                    ),
                ),
                (
                    "addr_shipping_address_1",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=250,
                        verbose_name="Adresszeile 1",
                    ),
                ),
                (
                    "addr_shipping_address_2",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=250,
                        verbose_name="Adresszeile 2",
                    ),
                ),
                (
                    "addr_shipping_city",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Ort"
                    ),
                ),
                (
                    "addr_shipping_state",
                    models.CharField(
                        blank=True, default="", max_length=50, verbose_name="Kanton"
                    ),
                ),
                (
                    "addr_shipping_postcode",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=50,
                        verbose_name="Postleitzahl",
                    ),
                ),
                (
                    "addr_shipping_country",
                    models.CharField(
                        choices=[("CH", "Schweiz"), ("LI", "Liechtenstein")],
                        default="CH",
                        max_length=2,
                        verbose_name="Land",
                    ),
                ),
                (
                    "addr_shipping_email",
                    models.EmailField(
                        blank=True, max_length=254, verbose_name="E-Mail-Adresse"
                    ),
                ),
                (
                    "addr_shipping_phone",
                    models.CharField(
                        blank=True, default="", max_length=50, verbose_name="Telefon"
                    ),
                ),
                (
                    "cached_sum",
                    models.FloatField(default=0.0, verbose_name="Total in CHF"),
                ),
                (
                    "pdf_title",
                    models.CharField(
                        blank=True, default="", editable=False, max_length=32
                    ),
                ),
                ("pdf_text", models.TextField(blank=True, default="", editable=False)),
                (
                    "contact_person",
                    models.ForeignKey(
                        default=kmuhelper.modules.main.models.default_contact_person,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="kmuhelper.contactperson",
                        verbose_name="Ansprechpartner",
                    ),
                ),
                (
                    "customer",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="orders",
                        to="kmuhelper.customer",
                        verbose_name="Kunde",
                    ),
                ),
                (
                    "email_link_invoice",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="kmuhelper.email",
                    ),
                ),
                (
                    "email_link_shipped",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="kmuhelper.email",
                    ),
                ),
            ],
            options={
                "verbose_name": "Bestellung",
                "verbose_name_plural": "Bestellungen",
            },
        ),
        migrations.CreateModel(
            name="PaymentImport",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "time_imported",
                    models.DateTimeField(
                        auto_now_add=True, verbose_name="Importiert am"
                    ),
                ),
                (
                    "is_processed",
                    models.BooleanField(default=False, verbose_name="Verarbeitet?"),
                ),
                (
                    "data_msgid",
                    models.CharField(max_length=50, verbose_name="Nachrichtenid"),
                ),
                ("data_creationdate", models.DateTimeField(verbose_name="Erstellt am")),
            ],
            options={
                "verbose_name": "Zahlungsimport",
                "verbose_name_plural": "Zahlungsimporte",
            },
        ),
        migrations.CreateModel(
            name="PaymentReceiver",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "mode",
                    models.CharField(
                        choices=[("QRR", "QR-Referenz"), ("NON", "Ohne Referenz")],
                        default="QRR",
                        max_length=15,
                        verbose_name="Modus",
                    ),
                ),
                (
                    "qriban",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="QR-IBAN mit Leerzeichen (Nur verwendet im Modus 'QR-Referenz')",
                        max_length=26,
                        validators=[
                            django.core.validators.RegexValidator(
                                "^(CH|LI)[0-9]{2}\\s3[0-9]{3}\\s[0-9]{4}\\s[0-9]{4}\\s[0-9]{4}\\s[0-9]{1}$",
                                "Bite benutze folgendes Format: (CH|LI)pp 3xxx xxxx xxxx xxxx x",
                            )
                        ],
                        verbose_name="QR-IBAN",
                    ),
                ),
                (
                    "iban",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="IBAN mit Leerzeichen (Nur verwendet im Modus 'Ohne Referenz')",
                        max_length=26,
                        validators=[
                            django.core.validators.RegexValidator(
                                "^(CH|LI)[0-9]{2}\\s[0-9]{4}\\s[0-9]{4}\\s[0-9]{4}\\s[0-9]{4}\\s[0-9]{1}$",
                                "Bite benutze folgendes Format: (CH|LI)pp 3xxx xxxx xxxx xxxx x",
                            )
                        ],
                        verbose_name="IBAN",
                    ),
                ),
                (
                    "logourl",
                    models.URLField(
                        blank=True,
                        default="",
                        help_text="URL eines Bildes (.jpg/.png) - Wird auf die Rechnung gedruckt.",
                        validators=[
                            django.core.validators.RegexValidator(
                                "^[0-9a-zA-Z\\-\\.\\|\\?\\(\\)\\*\\+&\"'_:;/]+\\.(png|jpg)$",
                                "Nur folgende Zeichen gestattet: 0-9a-zA-Z-_.:;/|?&()\"'*+ - Muss auf .jpg/.png enden.",
                            )
                        ],
                        verbose_name="Logo (URL)",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Name der Firma oder des Empfängers",
                        max_length=70,
                        verbose_name="Name",
                    ),
                ),
                (
                    "swiss_uid",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="UID der Firma - Format: CHE-123.456.789 (Mehrwertsteuernummer)",
                        max_length=15,
                        validators=[
                            django.core.validators.RegexValidator(
                                "^CHE-[0-9]{3}\\.[0-9]{3}\\.[0-9]{3}$",
                                "Bite benutze folgendes Format: CHE-123.456.789",
                            )
                        ],
                        verbose_name="Firmen-UID",
                    ),
                ),
                (
                    "address_1",
                    models.CharField(
                        max_length=70,
                        verbose_name="Strasse und Hausnummer oder 'Postfach'",
                    ),
                ),
                (
                    "address_2",
                    models.CharField(max_length=70, verbose_name="PLZ und Ort"),
                ),
                (
                    "country",
                    models.CharField(
                        choices=[("CH", "Schweiz"), ("LI", "Liechtenstein")],
                        default="CH",
                        max_length=2,
                        verbose_name="Land",
                    ),
                ),
                (
                    "email",
                    models.EmailField(
                        blank=True,
                        default="",
                        help_text="Nicht auf der Rechnung ersichtlich",
                        max_length=254,
                        verbose_name="E-Mail",
                    ),
                ),
                (
                    "phone",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Nicht auf der Rechnung ersichtlich",
                        max_length=70,
                        verbose_name="Telefon",
                    ),
                ),
                (
                    "website",
                    models.URLField(
                        blank=True,
                        default="",
                        help_text="Auf der Rechnung ersichtlich, sofern vorhanden!",
                        verbose_name="Webseite",
                    ),
                ),
            ],
            options={
                "verbose_name": "Zahlungsempfänger",
                "verbose_name_plural": "Zahlungsempfänger",
            },
        ),
        migrations.CreateModel(
            name="Product",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "article_number",
                    models.CharField(max_length=25, verbose_name="Artikelnummer"),
                ),
                (
                    "woocommerceid",
                    models.IntegerField(default=0, verbose_name="WooCommerce ID"),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Unterstützt i18n: '[:de]Deutsch[:fr]Français[:it]Italiano[:en]English[:]'",
                        max_length=500,
                        verbose_name="Name",
                    ),
                ),
                (
                    "short_description",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="Unterstützt i18n: '[:de]Deutsch[:fr]Français[:it]Italiano[:en]English[:]'",
                        verbose_name="Kurzbeschrieb",
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="Unterstützt i18n: '[:de]Deutsch[:fr]Français[:it]Italiano[:en]English[:]'",
                        verbose_name="Beschrieb",
                    ),
                ),
                (
                    "quantity_description",
                    models.CharField(
                        blank=True,
                        default="Stück",
                        help_text="Unterstützt i18n: '[:de]Deutsch[:fr]Français[:it]Italiano[:en]English[:]'",
                        max_length=100,
                        verbose_name="Mengenbezeichnung",
                    ),
                ),
                (
                    "selling_price",
                    models.FloatField(
                        default=0, verbose_name="Normalpreis in CHF (exkl. MwSt)"
                    ),
                ),
                (
                    "vat_rate",
                    models.FloatField(
                        choices=[
                            (0.0, "0.0% (Mehrwertsteuerfrei)"),
                            (2.5, "2.5% (Bis 2023: Reduzierter Satz)"),
                            (2.6, "2.6% (Ab 2024: Reduzierter Satz)"),
                            (3.7, "3.7% (Bis 2023: Sondersatz für Beherbergung)"),
                            (3.8, "3.8% (Ab 2024: Sondersatz für Beherbergung)"),
                            (7.7, "7.7% (Bis 2023: Normalsatz)"),
                            (8.1, "8.1% (Ab 2024: Normalsatz)"),
                        ],
                        default=7.7,
                        verbose_name="MwSt-Satz",
                    ),
                ),
                (
                    "stock_current",
                    models.IntegerField(default=0, verbose_name="Lagerbestand"),
                ),
                (
                    "stock_target",
                    models.IntegerField(default=1, verbose_name="Soll-Lagerbestand"),
                ),
                (
                    "note",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="Wird nicht gedruckt oder angezeigt; nur für eigene Zwecke.",
                        verbose_name="Bemerkung",
                    ),
                ),
                (
                    "sale_from",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="In Aktion von"
                    ),
                ),
                (
                    "sale_to",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="In Aktion bis"
                    ),
                ),
                (
                    "sale_price",
                    models.FloatField(
                        blank=True,
                        null=True,
                        verbose_name="Aktionspreis in CHF (exkl. MwSt)",
                    ),
                ),
                (
                    "datasheet_url",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=500,
                        verbose_name="Datenblattlink",
                    ),
                ),
                (
                    "image_url",
                    models.URLField(blank=True, default="", verbose_name="Bildlink"),
                ),
                (
                    "supplier_price",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=20,
                        verbose_name="Einkaufspreis",
                    ),
                ),
                (
                    "supplier_article_number",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=25,
                        verbose_name="Lieferantenartikelnummer",
                    ),
                ),
                (
                    "supplier_url",
                    models.URLField(
                        blank=True,
                        default="",
                        verbose_name="Lieferantenurl (Für Nachbestellungen)",
                    ),
                ),
            ],
            options={
                "verbose_name": "Produkt",
                "verbose_name_plural": "Produkte",
            },
        ),
        migrations.CreateModel(
            name="ProductCategory",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "woocommerceid",
                    models.IntegerField(default=0, verbose_name="WooCommerce ID"),
                ),
                (
                    "name",
                    models.CharField(
                        default="", max_length=250, verbose_name="Bezeichnung"
                    ),
                ),
                (
                    "description",
                    models.TextField(blank=True, default="", verbose_name="Beschrieb"),
                ),
                ("image_url", models.URLField(blank=True, verbose_name="Bildlink")),
                (
                    "parent_category",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="kmuhelper.productcategory",
                        verbose_name="Übergeordnete Kategorie",
                    ),
                ),
            ],
            options={
                "verbose_name": "Produktkategorie",
                "verbose_name_plural": "Produktkategorien",
            },
        ),
        migrations.CreateModel(
            name="Setting",
            fields=[
                (
                    "id",
                    models.CharField(
                        max_length=50,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "typ",
                    models.CharField(
                        choices=[
                            ("char", "Text"),
                            ("text", "Mehrzeiliger Text"),
                            ("bool", "Wahrheitswert"),
                            ("int", "Zahl"),
                            ("float", "Fliesskommazahl"),
                            ("url", "Url"),
                            ("email", "E-Mail"),
                            ("json", "JSON Daten"),
                        ],
                        default="char",
                        max_length=5,
                        verbose_name="Typ",
                    ),
                ),
                (
                    "content_char",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=250,
                        verbose_name="Inhalt (Text)",
                    ),
                ),
                (
                    "content_text",
                    models.TextField(
                        blank=True,
                        default="",
                        verbose_name="Inhalt (Mehrzeiliger Text)",
                    ),
                ),
                (
                    "content_bool",
                    models.BooleanField(
                        blank=True, default=False, verbose_name="Inhalt (Wahrheitswert)"
                    ),
                ),
                (
                    "content_int",
                    models.IntegerField(
                        blank=True, default=0, verbose_name="Inhalt (Zahl)"
                    ),
                ),
                (
                    "content_float",
                    models.FloatField(
                        blank=True, default=0.0, verbose_name="Inhalt (Fliesskommazahl)"
                    ),
                ),
                (
                    "content_url",
                    models.URLField(
                        blank=True, default="", verbose_name="Inhalt (Url)"
                    ),
                ),
                (
                    "content_email",
                    models.EmailField(
                        blank=True,
                        default="",
                        max_length=254,
                        verbose_name="Inhalt (E-Mail)",
                    ),
                ),
                (
                    "content_json",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        null=True,
                        verbose_name="Inhalt (JSON)",
                    ),
                ),
            ],
            options={
                "verbose_name": "Einstellung",
                "verbose_name_plural": "Einstellungen",
            },
        ),
        migrations.CreateModel(
            name="SettingHidden",
            fields=[
                (
                    "id",
                    models.CharField(
                        max_length=50,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "typ",
                    models.CharField(
                        choices=[
                            ("char", "Text"),
                            ("text", "Mehrzeiliger Text"),
                            ("bool", "Wahrheitswert"),
                            ("int", "Zahl"),
                            ("float", "Fliesskommazahl"),
                            ("url", "Url"),
                            ("email", "E-Mail"),
                            ("json", "JSON Daten"),
                        ],
                        default="char",
                        max_length=5,
                        verbose_name="Typ",
                    ),
                ),
                (
                    "content_char",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=250,
                        verbose_name="Inhalt (Text)",
                    ),
                ),
                (
                    "content_text",
                    models.TextField(
                        blank=True,
                        default="",
                        verbose_name="Inhalt (Mehrzeiliger Text)",
                    ),
                ),
                (
                    "content_bool",
                    models.BooleanField(
                        blank=True, default=False, verbose_name="Inhalt (Wahrheitswert)"
                    ),
                ),
                (
                    "content_int",
                    models.IntegerField(
                        blank=True, default=0, verbose_name="Inhalt (Zahl)"
                    ),
                ),
                (
                    "content_float",
                    models.FloatField(
                        blank=True, default=0.0, verbose_name="Inhalt (Fliesskommazahl)"
                    ),
                ),
                (
                    "content_url",
                    models.URLField(
                        blank=True, default="", verbose_name="Inhalt (Url)"
                    ),
                ),
                (
                    "content_email",
                    models.EmailField(
                        blank=True,
                        default="",
                        max_length=254,
                        verbose_name="Inhalt (E-Mail)",
                    ),
                ),
                (
                    "content_json",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        null=True,
                        verbose_name="Inhalt (JSON)",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Supplier",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("abbreviation", models.CharField(max_length=5, verbose_name="Kürzel")),
                ("name", models.CharField(max_length=50, verbose_name="Name")),
                ("website", models.URLField(blank=True, verbose_name="Webseite")),
                (
                    "phone",
                    models.CharField(
                        blank=True, default="", max_length=50, verbose_name="Telefon"
                    ),
                ),
                (
                    "email",
                    models.EmailField(
                        blank=True, max_length=254, null=True, verbose_name="E-Mail"
                    ),
                ),
                (
                    "address",
                    models.TextField(blank=True, default="", verbose_name="Adresse"),
                ),
                (
                    "note",
                    models.TextField(blank=True, default="", verbose_name="Notiz"),
                ),
                (
                    "contact_person_name",
                    models.CharField(
                        blank=True, default="", max_length=250, verbose_name="Name"
                    ),
                ),
                (
                    "contact_person_phone",
                    models.CharField(
                        blank=True, default="", max_length=50, verbose_name="Telefon"
                    ),
                ),
                (
                    "contact_person_email",
                    models.EmailField(
                        blank=True,
                        default="",
                        max_length=254,
                        null=True,
                        verbose_name="E-Mail",
                    ),
                ),
            ],
            options={
                "verbose_name": "Lieferant",
                "verbose_name_plural": "Lieferanten",
            },
        ),
        migrations.CreateModel(
            name="Supply",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        default=kmuhelper.modules.main.models.default_delivery_title,
                        max_length=50,
                        verbose_name="Name",
                    ),
                ),
                (
                    "date",
                    models.DateField(auto_now_add=True, verbose_name="Erfasst am"),
                ),
                (
                    "is_added_to_stock",
                    models.BooleanField(default=False, verbose_name="Eingelagert?"),
                ),
            ],
            options={
                "verbose_name": "Lieferung",
                "verbose_name_plural": "Lieferungen",
            },
        ),
        migrations.CreateModel(
            name="SupplyItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("quantity", models.IntegerField(default=1, verbose_name="Menge")),
                (
                    "product",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="kmuhelper.product",
                        verbose_name="Produkt",
                    ),
                ),
                (
                    "supply",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="kmuhelper.supply",
                    ),
                ),
            ],
            options={
                "verbose_name": "Lieferungsposten",
                "verbose_name_plural": "Lieferungsposten",
            },
        ),
        migrations.AddField(
            model_name="supply",
            name="products",
            field=models.ManyToManyField(
                through="kmuhelper.SupplyItem", to="kmuhelper.product"
            ),
        ),
        migrations.AddField(
            model_name="supply",
            name="supplier",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="kmuhelper.supplier",
                verbose_name="Lieferant",
            ),
        ),
        migrations.CreateModel(
            name="ProductProductCategoryConnection",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "category",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="kmuhelper.productcategory",
                        verbose_name="Produktkategorie",
                    ),
                ),
                (
                    "product",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="kmuhelper.product",
                        verbose_name="Produkt",
                    ),
                ),
            ],
            options={
                "verbose_name": "Produkt-Kategorie-Verknüpfung",
                "verbose_name_plural": "Produkt-Kategorie-Verknüpfungen",
            },
        ),
        migrations.AddField(
            model_name="product",
            name="categories",
            field=models.ManyToManyField(
                related_name="products",
                through="kmuhelper.ProductProductCategoryConnection",
                to="kmuhelper.productcategory",
                verbose_name="Kategorien",
            ),
        ),
        migrations.AddField(
            model_name="product",
            name="supplier",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="kmuhelper.supplier",
                verbose_name="Lieferant",
            ),
        ),
        migrations.CreateModel(
            name="PaymentImportEntry",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "ref",
                    models.CharField(
                        default="", max_length=50, verbose_name="Referenznummer"
                    ),
                ),
                (
                    "additionalref",
                    models.CharField(
                        default="", max_length=250, verbose_name="Zusätzliche Referenz"
                    ),
                ),
                (
                    "iban",
                    models.CharField(default="", max_length=22, verbose_name="IBAN"),
                ),
                (
                    "name",
                    models.CharField(default="", max_length=250, verbose_name="Name"),
                ),
                ("valuedate", models.DateField(null=True, verbose_name="Valuta")),
                ("amount", models.FloatField(verbose_name="Betrag")),
                ("currency", models.CharField(max_length=10, verbose_name="Währung")),
                (
                    "parent",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="entries",
                        to="kmuhelper.paymentimport",
                    ),
                ),
            ],
            options={
                "verbose_name": "Zahlungsimport-Eintrag",
                "verbose_name_plural": "Zahlungsimport-Einträge",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="OrderItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "note",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Wird auf die Rechnung gedruckt.",
                        max_length=250,
                        verbose_name="Bemerkung",
                    ),
                ),
                ("quantity", models.IntegerField(default=1, verbose_name="Menge")),
                (
                    "discount",
                    models.FloatField(
                        default=0,
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(100),
                        ],
                        verbose_name="Rabatt in %",
                    ),
                ),
                (
                    "product_price",
                    models.FloatField(
                        default=0.0, verbose_name="Produktpreis (exkl. MwSt)"
                    ),
                ),
                (
                    "order",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="kmuhelper.order",
                    ),
                ),
                (
                    "product",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="kmuhelper.product",
                        verbose_name="Produkt",
                    ),
                ),
            ],
            options={
                "verbose_name": "Bestellungsposten",
                "verbose_name_plural": "Bestellungsposten",
            },
        ),
        migrations.CreateModel(
            name="OrderFee",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "note",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Wird auf die Rechnung gedruckt.",
                        max_length=250,
                        verbose_name="Bemerkung",
                    ),
                ),
                (
                    "discount",
                    models.FloatField(
                        default=0.0,
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(100),
                        ],
                        verbose_name="Rabatt in %",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        default="Zusätzliche Kosten",
                        help_text="Unterstützt i18n: '[:de]Deutsch[:fr]Français[:it]Italiano[:en]English[:]'",
                        max_length=500,
                        verbose_name="Name",
                    ),
                ),
                (
                    "price",
                    models.FloatField(default=0.0, verbose_name="Preis (exkl. MwSt)"),
                ),
                (
                    "vat_rate",
                    models.FloatField(
                        choices=[
                            (0.0, "0.0% (Mehrwertsteuerfrei)"),
                            (2.5, "2.5% (Bis 2023: Reduzierter Satz)"),
                            (2.6, "2.6% (Ab 2024: Reduzierter Satz)"),
                            (3.7, "3.7% (Bis 2023: Sondersatz für Beherbergung)"),
                            (3.8, "3.8% (Ab 2024: Sondersatz für Beherbergung)"),
                            (7.7, "7.7% (Bis 2023: Normalsatz)"),
                            (8.1, "8.1% (Ab 2024: Normalsatz)"),
                        ],
                        default=7.7,
                        verbose_name="MwSt-Satz",
                    ),
                ),
                (
                    "linked_fee",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="kmuhelper.fee",
                        verbose_name="Verknüpfte Kosten",
                    ),
                ),
                (
                    "order",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="kmuhelper.order",
                    ),
                ),
            ],
            options={
                "verbose_name": "Bestellungskosten",
                "verbose_name_plural": "Bestellungskosten",
            },
        ),
        migrations.AddField(
            model_name="order",
            name="fees",
            field=models.ManyToManyField(
                through="kmuhelper.OrderFee", to="kmuhelper.fee"
            ),
        ),
        migrations.AddField(
            model_name="order",
            name="payment_receiver",
            field=models.ForeignKey(
                default=kmuhelper.modules.main.models.default_payment_recipient,
                on_delete=django.db.models.deletion.PROTECT,
                to="kmuhelper.paymentreceiver",
                verbose_name="Zahlungsempfänger",
            ),
        ),
        migrations.AddField(
            model_name="order",
            name="products",
            field=models.ManyToManyField(
                through="kmuhelper.OrderItem", to="kmuhelper.product"
            ),
        ),
        migrations.CreateModel(
            name="Note",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=50, verbose_name="Name")),
                (
                    "description",
                    models.TextField(blank=True, default="", verbose_name="Beschrieb"),
                ),
                ("done", models.BooleanField(default=False, verbose_name="Erledigt?")),
                (
                    "priority",
                    models.IntegerField(
                        blank=True, default=0, verbose_name="Priorität"
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="Erstellt am"),
                ),
                (
                    "linked_customer",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="linked_note",
                        to="kmuhelper.customer",
                    ),
                ),
                (
                    "linked_order",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="linked_note",
                        to="kmuhelper.order",
                    ),
                ),
                (
                    "linked_product",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="linked_note",
                        to="kmuhelper.product",
                    ),
                ),
                (
                    "linked_supply",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="linked_note",
                        to="kmuhelper.supply",
                    ),
                ),
            ],
            options={
                "verbose_name": "Notiz",
                "verbose_name_plural": "Notizen",
            },
        ),
        migrations.CreateModel(
            name="EMailAttachment",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "attachment",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="emails",
                        to="kmuhelper.attachment",
                    ),
                ),
                (
                    "email",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="kmuhelper.email",
                    ),
                ),
            ],
            options={
                "verbose_name": "E-Mail-Anhang",
                "verbose_name_plural": "E-Mail-Anhänge",
            },
        ),
        migrations.AddField(
            model_name="email",
            name="attachments",
            field=models.ManyToManyField(
                through="kmuhelper.EMailAttachment", to="kmuhelper.attachment"
            ),
        ),
        migrations.AddField(
            model_name="customer",
            name="email_link_registered",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="kmuhelper.email",
            ),
        ),
        migrations.CreateModel(
            name="ApiKey",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "key",
                    models.UUIDField(
                        default=uuid.uuid4, unique=True, verbose_name="Key"
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        blank=True, default="", max_length=100, verbose_name="Name"
                    ),
                ),
                (
                    "read",
                    models.BooleanField(default=True, verbose_name="Read permission?"),
                ),
                (
                    "write",
                    models.BooleanField(
                        default=False, verbose_name="Write permission?"
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "Api key",
                "verbose_name_plural": "Api keys",
            },
        ),
        migrations.CreateModel(
            name="App_Arrival",
            fields=[],
            options={
                "verbose_name": "Lieferung",
                "verbose_name_plural": "Lieferungen",
                "proxy": True,
                "default_permissions": ("add", "change", "view"),
                "indexes": [],
                "constraints": [],
            },
            bases=("kmuhelper.supply",),
        ),
        migrations.CreateModel(
            name="App_IncomingPayments",
            fields=[],
            options={
                "verbose_name": "Bestellung",
                "verbose_name_plural": "Bestellungen",
                "proxy": True,
                "default_permissions": ("add", "change", "view"),
                "indexes": [],
                "constraints": [],
            },
            bases=("kmuhelper.order",),
        ),
        migrations.CreateModel(
            name="App_Shipping",
            fields=[],
            options={
                "verbose_name": "Bestellung",
                "verbose_name_plural": "Bestellungen",
                "proxy": True,
                "default_permissions": ("add", "change", "view"),
                "indexes": [],
                "constraints": [],
            },
            bases=("kmuhelper.order",),
        ),
        migrations.CreateModel(
            name="App_Stock",
            fields=[],
            options={
                "verbose_name": "Produkt",
                "verbose_name_plural": "Produkte",
                "proxy": True,
                "default_permissions": ("add", "change", "view"),
                "indexes": [],
                "constraints": [],
            },
            bases=("kmuhelper.product",),
        ),
        migrations.CreateModel(
            name="App_ToDo",
            fields=[],
            options={
                "verbose_name": "Notiz",
                "verbose_name_plural": "Notizen",
                "proxy": True,
                "default_permissions": ("add", "change", "view"),
                "indexes": [],
                "constraints": [],
            },
            bases=("kmuhelper.note",),
        ),
    ]



from defusedxml.ElementTree import parse as parse_xml, ParseError

from rich import print

from django import forms
from django.contrib import messages
from django.core.validators import FileExtensionValidator
from django.shortcuts import redirect, reverse

from django.utils.translation import gettext_lazy, gettext

from kmuhelper.modules.integrations.paymentimport.models import (
    PaymentImport,
    PaymentImportEntry,
)

_ = gettext_lazy


def log(string, *args):
    print("[deep_pink4][KMUHelper Paymentimport][/] -", string, *args)


#####


class CamtUploadForm(forms.Form):
    """Form for uploading and processing camt.054.001.02 files"""

    # Fieldsets

    fieldsets = [{"fields": ["file"], "name": _("Dateiupload")}]

    # Fields

    file = forms.FileField(
        label=_("XML-Datei"),
        help_text=_(
            "Die Datei muss nach dem camt.053.001.04 Standard aufgebaut sein und Detailinformationen enthalten."
        ),
        validators=[FileExtensionValidator(["xml"])],
    )

    def clean_file(self):
        file = self.cleaned_data.get("file")

        if file:
            try:
                tree = parse_xml(file)
                root = tree.getroot()
                msg = root.find("{*}BkToCstmrStmt")

                if "camt.053.001.04" not in str(root.tag) or msg is None:
                    raise forms.ValidationError(
                        gettext("Die Datei liegt nicht im korrekten Format vor!")
                    )

                self.cleaned_data["camt.053.001.04"] = msg
            except ParseError as error:
                raise forms.ValidationError(
                    gettext("Die Datei kann nicht verarbeitet werden! Fehler: %s")
                    % error
                )

        return file

    def create_payment_import_and_redirect(self, request):
        msg = self.cleaned_data.get("camt.053.001.04")
        try:
            msgid = msg.find("./{*}GrpHdr/{*}MsgId")
            msgid = msgid.text if msgid is not None else ""

            if msgid and PaymentImport.objects.filter(data_msgid=msgid).exists():
                dbentry = PaymentImport.objects.get(data_msgid=msgid)
                messages.error(
                    request,
                    _(
                        "Eine Datei mit der Nachrichtenid '%s' wurde bereits zu einem früheren Zeitpunkt importiert! Der Import wurde abgebrochen!"
                    )
                    % msgid,
                )
                return redirect(
                    reverse("admin:kmuhelper_paymentimport_change", args=[dbentry.pk])
                )

            creationdate = msg.find("./{*}GrpHdr/{*}CreDtTm").text
            log(
                f"Start processing - Message id: '{msgid}' - Creation date: '{creationdate}'"
            )

            dbentry = PaymentImport.objects.create(
                data_msgid=msgid,
                data_creationdate=creationdate,
            )

            for entry in msg.findall("./{*}Stmt/{*}Ntry"):
                description = entry.find("{*}AddtlNtryInf").text
                iscredit = entry.find("{*}CdtDbtInd").text == "CRDT"

                valuedate = entry.find("./{*}ValDt/{*}Dt")
                valuedate = valuedate.text if valuedate is not None else None

                log(
                    f"- Entry {iscredit} Description: '{description}' Value: {valuedate}"
                )
                if iscredit:
                    if description in ["Sammelbuchung QR-Rechnung mit QR-Referenz"]:
                        for txdtls in entry.findall("./{*}NtryDtls/{*}TxDtls"):
                            _amt = txdtls.find("{*}Amt")
                            amount = _amt.text
                            currency = _amt.attrib.get("Ccy", "CHF")

                            name = txdtls.find("./{*}RltdPties/{*}Dbtr/{*}Nm")
                            name = name.text if name is not None else ""
                            iban = txdtls.find(
                                "./{*}RltdPties/{*}DbtrAcct/{*}Id/{*}IBAN"
                            )
                            iban = iban.text if iban is not None else ""
                            ref = txdtls.find(
                                "./{*}RmtInf/{*}Strd/{*}CdtrRefInf/{*}Ref"
                            )
                            ref = ref.text if ref is not None else ""
                            addref = txdtls.find("./{*}RmtInf/{*}Strd/{*}AddtlRmtInf")
                            addref = addref.text if addref is not None else ""

                            log(
                                f"- - {amount} {currency} '{name}' '{iban}' {ref} '{addref}'"
                            )

                            PaymentImportEntry.objects.create(
                                parent=dbentry,
                                name=name,
                                iban=iban,
                                ref=ref,
                                additionalref=addref,
                                valuedate=valuedate,
                                amount=float(amount),
                                currency=currency,
                            )

            messages.success(request, _("Zahlungsimport wurde erfolgreich importiert!"))
            return redirect(
                reverse("admin:kmuhelper_paymentimport_process", args=[dbentry.pk])
            )
        except AttributeError as error:
            log(error)
            messages.error(
                request,
                _("Bei der Verarbeitung ist ein Fehler aufgetreten: %s") % error,
            )
            return redirect(reverse("admin:kmuhelper_paymentimport_upload"))



import uuid

from django.contrib import admin, messages
from django.core import mail
from django.core.exceptions import ValidationError
from django.core.files import storage
from django.db import models
from django.http import FileResponse
from django.template import TemplateDoesNotExist, TemplateSyntaxError, Template, Context
from django.template.loader import get_template
from django.urls import reverse
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import gettext_lazy, gettext
from kmuhelper import settings, constants
from kmuhelper.external.multi_email_field.fields import MultiEmailField
from kmuhelper.overrides import CustomModel
from kmuhelper.translations import Language
from rich import print

_ = gettext_lazy


def log(string, *args):
    print("[deep_pink4][KMUHelper E-Mails][/] -", string, *args)


#####


def getfilepath(instance, filename):
    return "attachments/" + timezone.now().strftime("%Y-%m-%d_%H-%M-%S_") + filename


class AttachmentManager(models.Manager):
    """Manager for the 'Attachment' model"""

    def create_from_binary(self, filename, content):
        """Create an Attachment object from binary data"""

        filepath = getfilepath(None, filename)

        file = storage.default_storage.save(filepath, content)
        return self.create(
            filename=filename,
            file=file,
            autocreated=True,
        )


class Attachment(CustomModel):
    """Model representing an attachment"""

    filename = models.CharField(
        verbose_name=_("Dateiname"),
        max_length=50,
    )
    file = models.FileField(
        verbose_name=_("Datei"),
        upload_to=getfilepath,
        storage=storage.default_storage,
    )

    description = models.TextField(
        verbose_name=_("Beschreibung"),
        default="",
        blank=True,
    )
    autocreated = models.BooleanField(
        verbose_name=_("Automatisch generiert"),
        default=False,
    )

    token = models.UUIDField(
        verbose_name=_("Token"),
        default=uuid.uuid4,
        editable=False,
    )

    time_created = models.DateTimeField(
        verbose_name=_("Erstellt um"),
        auto_now_add=True,
    )

    @admin.display(description=_("Anhang"))
    def __str__(self):
        return f"{self.filename} ({self.pk})"

    def get_file_response(self, download=False):
        """Get this attachment as a file response"""

        return FileResponse(
            storage.default_storage.open(self.file.path, "rb"),
            as_attachment=download,
            filename=self.filename,
        )

    def get_url(self):
        """Get the public view online URL"""

        path = reverse("kmuhelper:attachment-view", kwargs={"object_id": self.pk})
        return path + f"?token={self.token}"

    class Meta:
        verbose_name = _("Anhang")
        verbose_name_plural = _("Anhänge")
        default_permissions = ("add", "change", "view", "delete", "download")

    objects = AttachmentManager()

    ADMIN_ICON = "fa-solid fa-paperclip"

    # Custom delete

    def delete(self, *args, **kwargs):
        """Delete the associated file before deleting the model object"""

        self.file.delete()
        super().delete(*args, **kwargs)


class EMailAttachment(CustomModel):
    """Model representing the connection between 'EMail' and 'Attachment'"""

    attachment = models.ForeignKey(
        to="Attachment",
        on_delete=models.PROTECT,
        related_name="emails",
    )
    email = models.ForeignKey(
        to="EMail",
        on_delete=models.CASCADE,
    )

    @admin.display(description=_("E-Mail-Anhang"))
    def __str__(self):
        return f"({self.email.pk}) -> {self.attachment}"

    class Meta:
        verbose_name = _("E-Mail-Anhang")
        verbose_name_plural = _("E-Mail-Anhänge")

    objects = models.Manager()


class EMail(CustomModel):
    """Model representing an email"""

    subject = models.CharField(
        verbose_name=_("Betreff"),
        max_length=50,
        help_text="Wird Standardmässig auch als Untertitel verwendet.",
    )

    to = MultiEmailField(
        verbose_name=_("Empfänger"),
        help_text="Direkte Empfänger",
    )
    cc = MultiEmailField(
        verbose_name=_("CC"),
        default="",
        blank=True,
        help_text=_("Kopie"),
    )
    bcc = MultiEmailField(
        verbose_name=_("BCC"),
        default="",
        blank=True,
        help_text=_("Blindkopie"),
    )

    language = models.CharField(
        verbose_name=_("Sprache"),
        default="de",
        choices=constants.LANGUAGES,
        max_length=2,
    )

    html_template = models.CharField(
        verbose_name=_("Designvorlage"),
        default="default.html",
        max_length=50,
        help_text=_("Dateiname der Designvorlage unter 'kmuhelper/emails/'."),
    )
    text = models.TextField(
        verbose_name=_("Text"),
        default="",
        blank=True,
        help_text=_(
            "Hauptinhalt - wird nicht von allen Designvorlagen verwendet. Links und E-Mail Adressen werden "
            "automatisch verlinkt."
        ),
    )
    html_context = models.JSONField(
        verbose_name=_("Daten"),
        default=dict,
        blank=True,
        null=True,
        help_text=_(
            "Daten im JSON-Format, mit welchen die Designvorlage befüllt wird."
        ),
    )

    token = models.UUIDField(
        verbose_name=_("Token"),
        default=uuid.uuid4,
        editable=False,
    )

    time_created = models.DateTimeField(
        verbose_name=_("Erstellt am"),
        auto_now_add=True,
        help_text=_("Datum und Zeit der Erstellung dieser E-Mail."),
    )
    time_sent = models.DateTimeField(
        verbose_name=_("Gesendet um"),
        blank=True,
        null=True,
        default=None,
        help_text=_("Datum und Zeit des letzten erfolgreichen Sendeversuches."),
    )

    sent = models.BooleanField(
        verbose_name="Gesendet?",
        default=False,
    )

    notes = models.TextField(
        verbose_name="Notizen",
        blank=True,
        default="",
        help_text=_("Diese Notizen haben keine Einwirkung auf die E-Mail selbst."),
    )

    attachments = models.ManyToManyField(
        to="Attachment",
        through="EMailAttachment",
    )

    @admin.display(description=_("E-Mail"))
    def __str__(self):
        return f"{self.subject} ({self.pk})"

    def is_valid(self, request=None):
        """Check if the email is valid.
        If not, add messages to request.

        Returns:
            0: Has errors
            1: Has warnings
            2: No warnings or errors
        """

        template = f"kmuhelper/emails/{self.html_template}"
        errors = []
        warnings = []

        # Check 1: Template

        try:
            get_template(template)
        except TemplateDoesNotExist:
            errors.append(gettext("Vorlage '%s' wurde nicht gefunden.") % template)
        except TemplateSyntaxError as error:
            errors.append(
                gettext("Vorlage '%(template)s' enthält ungültige Syntax: %(error)s")
                % {
                    "template": template,
                    "error": error,
                }
            )

        # Check 2: Receiver

        if not self.to:
            warnings.append(gettext("Nachricht hat keine(n) Empfänger!"))

        # Add messages and return

        if request:
            for msg in errors:
                messages.error(request, msg)
            for msg in warnings:
                messages.warning(request, msg)

        haserrors = len(errors) > 0
        haswarnings = len(warnings) > 0
        return 0 if haserrors else 1 if haswarnings else 2

    def get_context(self):
        """Get the context for rendering"""

        ctx = dict(self.html_context) if self.html_context is not None else {}

        defaultcontext = settings.get_file_setting(
            "KMUHELPER_EMAILS_DEFAULT_CONTEXT", {}
        )
        signature = settings.get_db_setting(
            "email-signature", ""
        ) or defaultcontext.get("postcontent", "")

        data = {
            **defaultcontext,
            "subtitle": self.subject,
            "postconent": signature,
            **ctx,
            "text": self.text,
        }

        return data

    def render(self, online=False):
        """Render the email and return the rendered string"""

        with Language(self.language):
            context = self.get_context()

            if online:
                context["isonline"] = True
                context["attachments"] = list(self.attachments.all())
            else:
                context["view_online_url"] = self.get_url_with_domain()
            return get_template("kmuhelper/emails/" + str(self.html_template)).render(
                context
            )

    def add_attachments(self, *attachments):
        """Add one or more Attachment objects to this EMail"""

        self.attachments.add(*attachments)
        self.save()

    def send(self, **options):
        """Send the mail with given attachments.
        Options are passed to django.core.mail.EmailMessage"""

        msg = mail.EmailMessage(
            subject=self.subject,
            body=self.render(),
            to=self.to.splitlines() if isinstance(self.to, str) else self.to,
            cc=self.cc.splitlines() if isinstance(self.cc, str) else self.cc,
            bcc=self.bcc.splitlines() if isinstance(self.bcc, str) else self.bcc,
            **options,
        )

        if settings.has_file_setting("KMUHELPER_LOG_EMAIL"):
            msg.bcc.append(settings.get_file_setting("KMUHELPER_LOG_EMAIL"))

        msg.content_subtype = "html"

        for attachment in self.attachments.all():
            msg.attach(filename=attachment.filename, content=attachment.file.read())

        success = msg.send()

        log("ID:", self.pk, "- Subject:", self.subject, "- Success:", success)

        if success:
            self.time_sent = timezone.now()
            self.sent = True
            self.save()
        return success

    def get_url(self):
        """Get the public view online URL"""

        path = reverse("kmuhelper:email-view", kwargs={"object_id": self.pk})
        return path + f"?token={self.token}"

    def get_url_with_domain(self):
        """Get the public view online URL with domain prefix"""

        domain = settings.get_file_setting("KMUHELPER_DOMAIN", None)

        if domain:
            url = self.get_url()
            return domain + url

        log(
            "[orange_red1]Einstellung KMUHELPER_DOMAIN ist nicht gesetzt! "
            + "E-Mails werden ohne 'online ansehen' Links versendet!"
        )
        return None

    class Meta:
        verbose_name = _("E-Mail")
        verbose_name_plural = _("E-Mails")
        default_permissions = ("add", "change", "view", "delete", "send")

    objects = models.Manager()

    ADMIN_ICON = "fa-solid fa-envelope"


class EMailTemplate(CustomModel):
    """Model representing an email template (not to confuse with design template)"""

    PKFILL_WIDTH = 3

    title = models.CharField(
        verbose_name=_("Titel"),
        max_length=50,
    )
    description = models.TextField(
        verbose_name=_("Beschreibung"),
        default="",
        blank=True,
    )

    mail_to = models.CharField(
        verbose_name=_("Empfänger"),
        max_length=50,
        default="",
        blank=True,
        help_text=_("Unterstützt Platzhalter. Siehe unten für mehr Informationen."),
    )
    mail_subject = models.CharField(
        verbose_name=_("Betreff"),
        max_length=50,
        help_text=_("Unterstützt Platzhalter. Siehe unten für mehr Informationen."),
    )
    mail_text = models.TextField(
        verbose_name=_("Text"),
        help_text=_("Unterstützt Platzhalter. Siehe unten für mehr Informationen."),
    )
    mail_template = models.CharField(
        verbose_name=_("Designvorlage"),
        default="default.html",
        max_length=50,
    )
    mail_context = models.JSONField(
        verbose_name=_("Daten"),
        default=dict,
        blank=True,
        null=True,
    )

    @admin.display(description=_("E-Mail-Vorlage"))
    def __str__(self):
        return f"{self.title} ({self.pk})"

    def clean(self):
        """Check if the template fields are valid"""

        errors = {}

        for field in ["mail_to", "mail_subject", "mail_text"]:
            try:
                Template(getattr(self, field))
            except TemplateSyntaxError as error:
                errors[field] = gettext(
                    "Vorlage enthält ungültige Syntax: %(error)s"
                ) % {
                    "error": error,
                }

        if errors:
            raise ValidationError(errors)

    @admin.display(description=_("Vorlage benutzen"))
    def get_use_link(self):
        link = reverse("admin:kmuhelper_emailtemplate_use", args=[self.pk])
        text = _("Vorlage benutzen")
        return format_html('<a href="{}">{}</a>', link, text)

    def create_mail(self, variables=None):
        """Use this template and replace variables"""

        if variables is None:
            variables = dict()

        def template_parse(template_string):
            template = Template(template_string)
            context = Context(variables)
            return template.render(context)

        email = EMail.objects.create(
            to=template_parse(self.mail_to),
            subject=template_parse(self.mail_subject),
            text=template_parse(self.mail_text),
            html_template=self.mail_template,
            html_context=self.mail_context,
        )

        return email

    class Meta:
        verbose_name = _("E-Mail-Vorlage")
        verbose_name_plural = _("E-Mail-Vorlagen")

    objects = models.Manager()

    ADMIN_ICON = "fa-solid fa-envelope-open-text"



from django.urls import path

from kmuhelper.modules.integrations.woocommerce import views

#######################

urlpatterns = [
    path("auth/start", views.wc_auth_start, name="wc-auth-start"),
    path("auth/end", views.wc_auth_end, name="wc-auth-end"),
    path("auth/key", views.wc_auth_key, name="wc-auth-key"),
    path("auth/clear", views.wc_auth_clear, name="wc-auth-clear"),
    path("import/products", views.wc_import_products, name="wc-import-products"),
    path("import/customers", views.wc_import_customers, name="wc-import-customers"),
    path("import/categories", views.wc_import_categories, name="wc-import-categories"),
    path("import/orders", views.wc_import_orders, name="wc-import-orders"),
    path(
        "update/product/<int:object_id>",
        views.wc_update_product,
        name="wc-update-product",
    ),
    path(
        "update/customer/<int:object_id>",
        views.wc_update_customer,
        name="wc-update-customer",
    ),
    path(
        "update/category/<int:object_id>",
        views.wc_update_category,
        name="wc-update-category",
    ),
    path("update/order/<int:object_id>", views.wc_update_order, name="wc-update-order"),
    path("webhooks", views.wc_webhooks, name="wc-webhooks"),
    # User Interface
    path("settings", views.wc_settings, name="wc-settings"),
    path("system-status", views.wc_system_status, name="wc-system-status"),
    path("webhooks-status", views.wc_webhooks_status, name="wc-webhooks-status"),
]



from django.contrib import admin
from django.urls import path
from django.utils.translation import gettext_lazy

from kmuhelper.modules.integrations.paymentimport import views
from kmuhelper.modules.integrations.paymentimport.models import (
    PaymentImport,
    PaymentImportEntry,
)
from kmuhelper.overrides import CustomModelAdmin

_ = gettext_lazy


class PaymentImportAdminEntryInline(admin.TabularInline):
    model = PaymentImportEntry
    verbose_name = _("Eintrag")
    verbose_name_plural = _("Einträge")
    extra = 0

    fields = (
        "currency",
        "betrag",
        "ref",
        "order_id",
        "valuedate",
        "name",
        "iban",
        "additionalref",
    )
    readonly_fields = (
        "order_id",
        "betrag",
    )
    ordering = ("ref",)


@admin.register(PaymentImport)
class PaymentImportAdmin(CustomModelAdmin):
    readonly_fields = ("time_imported",)

    fieldsets = [
        ("Infos", {"fields": ["is_processed", "time_imported"]}),
        ("Daten", {"fields": ["data_msgid", "data_creationdate"]}),
    ]

    ordering = ("is_processed",)

    list_display = (
        "pkfill",
        "time_imported",
        "entrycount",
        "is_processed",
        "data_msgid",
    )
    list_filter = ("is_processed",)

    inlines = (PaymentImportAdminEntryInline,)

    save_on_top = True

    # Views

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name

        urls = super().get_urls()

        my_urls = [
            path(
                "upload/",
                self.admin_site.admin_view(views.upload),
                name="%s_%s_upload" % info,
            ),
            path(
                "<path:object_id>/process/",
                self.admin_site.admin_view(views.process),
                name="%s_%s_process" % info,
            ),
        ]
        return my_urls + urls

    # Permissions

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return obj is None


#


modeladmins = [
    (PaymentImport, PaymentImportAdmin),
]



from django.utils.translation import pgettext
from reportlab.graphics import renderPDF
from reportlab.graphics.barcode import qr
from reportlab.graphics.shapes import Drawing
from reportlab.lib.units import mm
from reportlab.pdfgen.canvas import Canvas
from reportlab.platypus import Flowable

import kmuhelper.modules.main.mixins as mixins
import kmuhelper.modules.main.models as models
from kmuhelper.utils import formatprice


class QRInvoiceFlowable(Flowable):
    """A Flowable that draws a Swiss QR-Invoice."""

    def __init__(
        self,
        total,
        address,
        payment_receiver,
        billing_information,
        qr_reference_number="",
        unstructured_message="",
        add_cut_info=True,
    ):
        super().__init__()

        self.width = 210
        self.height = 110
        self._fixedWidth = 210
        self._fixedHeight = 110

        self.total: float = total
        self.address: mixins.Address = address
        self.payment_receiver: models.PaymentReceiver = payment_receiver
        self.billing_information: str = billing_information
        self.qr_reference_number: str = qr_reference_number
        self.unstructured_message: str = unstructured_message
        self.add_cut_info: bool = add_cut_info

    @classmethod
    def from_order(cls, order, add_cut_lines=True):
        elem = cls(
            total=order.cached_sum,
            address=order.addr_billing,
            payment_receiver=order.payment_receiver,
            billing_information=order.get_qr_billing_information(),
            qr_reference_number=order.get_qr_reference_number(),
            unstructured_message=order.get_unstructured_message(),
            add_cut_info=add_cut_lines,
        )

        return elem

    def __repr__(self):
        return "QR-Invoice"

    def __str__(self):
        return "QR-Invoice"

    def debug(self):
        c = self.canv
        c.setStrokeColor("green")
        c.rect(5 * mm, 5 * mm, 52 * mm, 95 * mm)  # Empfangsschein
        c.rect(67 * mm, 5 * mm, 138 * mm, 95 * mm)  # Zahlteil
        c.rect(67 * mm, 42 * mm, 46 * mm, 46 * mm)  # QR-Code
        c.line(5 * mm, 23 * mm, 57 * mm, 23 * mm)
        c.line(5 * mm, 37 * mm, 57 * mm, 37 * mm)
        c.line(5 * mm, 93 * mm, 57 * mm, 93 * mm)
        c.line(67 * mm, 93 * mm, 118 * mm, 93 * mm)
        c.line(67 * mm, 37 * mm, 118 * mm, 37 * mm)
        c.line(67 * mm, 15 * mm, 205 * mm, 15 * mm)
        c.line(118 * mm, 100 * mm, 118 * mm, 15 * mm)
        c.setStrokeColor("black")

    def get_swiss_qr_payload(self):
        recv = self.payment_receiver
        addr = self.address
        qrpayload = []

        def ln(text=""):
            qrpayload.append(text)

        # QRCH
        # - Header
        # - - QRType
        ln("SPC")
        # - - Version
        ln("0200")
        # - - Coding
        ln("1")

        # - CdtrInf (Empfänger)
        # - - IBAN
        if recv.mode == "QRR":
            ln(recv.qriban.replace(" ", ""))
        else:
            ln(recv.iban.replace(" ", ""))
        # - - Cdtr
        # - - - AdrTp
        ln("S")
        # - - - Name
        ln(recv.invoice_name)
        # - - - StrtNmOrAdrLine1
        ln(recv.invoice_street)
        # - - - BldgNbOrAdrLine2
        ln(recv.invoice_street_nr)
        # - - - PstCd
        ln(recv.invoice_postcode)
        # - - - TwnNm
        ln(recv.invoice_city)
        # - - - Ctry (2-stelliger Landescode gemäss ISO 3166-1)
        ln(recv.invoice_country)

        # - UltmtCdtr (Endgültiger Zahlungsempfänger)
        # - - AdrTp
        ln()
        # - - Name
        ln()
        # - - StrtNmOrAdrLine1
        ln()
        # - - BldgNbOrAdrLine2
        ln()
        # - - PstCd
        ln()
        # - - TwnNm
        ln()
        # - - Ctry (2-stelliger Landescode gemäss ISO 3166-1)
        ln()

        # - CcyAmt
        # - - Amt
        ln(formatprice(self.total) if self.total != 0 else "")
        # - - Ccy
        ln("CHF")

        # - UltmtDbtr (Endgültiger Zahlungspflichtiger)
        if addr.is_empty():
            ln("\n\n\n\n\n\n")
        else:
            # - - AdrTp
            ln("S")
            # - - Name
            ln(addr.company or f"{addr.first_name} {addr.last_name}")
            # - - StrtNmOrAdrLine1
            ln(addr.address_1)
            # - - BldgNbOrAdrLine2
            # Version 2.3 of the specification allows the building number to be delivered in StrtNmOrAdrLine1 (above)
            ln()
            # - - PstCd
            ln(addr.postcode)
            # - - TwnNm
            ln(addr.city)
            # - - Ctry (2-stelliger Landescode gemäss ISO 3166-1)
            ln(addr.country)

        # - RmtIn
        # - - TP
        ln(recv.mode)
        # - - Ref
        if recv.mode == "QRR":
            ln(self.qr_reference_number.replace(" ", ""))
        else:
            ln()
        # - - AddInf
        # - - - Ustrd
        ln(self.unstructured_message)
        # - - - Trailer
        ln("EPD")
        # - - - StrdBkgInf
        ln(self.billing_information)

        # - AltPmtInf
        # - - AltPmt
        # ln()
        # - - AltPmt
        # ln()

        return "\n".join(qrpayload)

    def draw_qr_invoice(self):
        ref = self.qr_reference_number
        recv = self.payment_receiver

        addr = self.address
        empty_addr = addr.is_empty()
        total = format(self.total, ",.2f").replace(",", " ")
        empty_total = self.total == 0

        c: Canvas = self.canv
        c.saveState()

        # QR-Code

        qrpayload = self.get_swiss_qr_payload()
        qr_code = qr.QrCodeWidget(qrpayload, barLevel="M")
        qrbounds = qr_code.getBounds()
        qrwidth = qrbounds[2] - qrbounds[0]
        qrheight = qrbounds[3] - qrbounds[1]
        d = Drawing(
            52.2 * mm,
            52.2 * mm,
            transform=[52.2 * mm / qrwidth, 0, 0, 52.2 * mm / qrheight, 0, 0],
        )  # 46, 46
        d.add(qr_code)
        renderPDF.draw(d, c, 63.9 * mm, 38.9 * mm)  # 67, 42

        # Schweizerkreuz

        c.setFillColor("black")
        c.setStrokeColor("white")
        c.rect(86.5 * mm, 61.5 * mm, 7 * mm, 7 * mm, fill=1, stroke=1)
        c.setFillColor("white")
        c.rect(89.25 * mm, 63 * mm, 1.5 * mm, 4 * mm, fill=1, stroke=0)
        c.rect(88 * mm, 64.25 * mm, 4 * mm, 1.5 * mm, fill=1, stroke=0)

        c.setFillColor("black")
        c.setStrokeColor("black")

        # Begrenzungen Empfangsschein und Zahlteil und Abzutrennen-Hinweis

        if self.add_cut_info:
            c.line(0 * mm, 105 * mm, 210 * mm, 105 * mm)
            c.line(62 * mm, 0 * mm, 62 * mm, 105 * mm)
            c.setFont("Helvetica-Bold", 8)
            c.drawCentredString(
                105 * mm,
                107 * mm,
                pgettext(
                    "QR-Invoice / fixed by SIX group style guide",
                    "Vor der Einzahlung abzutrennen",
                ),
            )

        # Common parts

        def _draw_corners(x, y, width, height):
            """Draw corners. X and y represent the top left coordinates."""
            COR_LEN = 3 * mm
            c.setLineWidth(0.75)

            p = c.beginPath()
            p.moveTo(x, y)  # top left
            p.lineTo(x + COR_LEN, y)
            p.moveTo(x + width - COR_LEN, y)
            p.lineTo(x + width, y)  # top right
            p.lineTo(x + width, y - COR_LEN)
            p.moveTo(x + width, y - height + COR_LEN)
            p.lineTo(x + width, y - height)  # bottom right
            p.lineTo(x + width - COR_LEN, y - height)
            p.moveTo(x + COR_LEN, y - height)
            p.lineTo(x, y - height)  # bottom left
            p.lineTo(x, y - height + COR_LEN)
            p.moveTo(x, y - COR_LEN)
            p.lineTo(x, y)  # top left

            c.drawPath(p, stroke=1, fill=0)

        def _write_title(_t, text, small=False):
            _t.setFont("Helvetica-Bold", 6 if small else 8)
            _t.textLine(text)
            _t.moveCursor(0, 2)
            _t.setFont("Helvetica", 8 if small else 10)

        def _write_creditor_and_reference(_t, small=False):
            _write_title(
                _t,
                pgettext(
                    "QR-Invoice / fixed by SIX group style guide", "Konto / Zahlbar an"
                ),
                small=small,
            )
            _t.textLine(recv.qriban if recv.mode == "QRR" else recv.iban)
            _t.textLine(recv.invoice_name)
            _t.textLine(f"{recv.invoice_street} {recv.invoice_street_nr}")
            _t.textLine(f"{recv.invoice_postcode} {recv.invoice_city}")
            _t.moveCursor(0, 9)
            if recv.mode == "QRR":
                _write_title(
                    _t,
                    pgettext("QR-Invoice / fixed by SIX group style guide", "Referenz"),
                    small=small,
                )
                _t.textLine(ref)
                _t.moveCursor(0, 9)

        def _write_additional_info(_t):
            _write_title(
                _t,
                pgettext(
                    "QR-Invoice / fixed by SIX group style guide",
                    "Zusätzliche Informationen",
                ),
            )
            _t.textLine(self.unstructured_message)
            _t.moveCursor(0, 9)

        def _write_debitor(_t, small=False):
            _write_title(
                _t,
                pgettext(
                    "QR-Invoice / fixed by SIX group style guide",
                    "Zahlbar durch (Name/Adresse)",
                )
                if empty_addr
                else pgettext(
                    "QR-Invoice / fixed by SIX group style guide", "Zahlbar durch"
                ),
                small=small,
            )
            if not empty_addr:
                _t.textLine(
                    addr["company"] or f"{addr['first_name']} {addr['last_name']}"
                )
                _t.textLine(addr["address_1"])
                _t.textLine(f"{addr['postcode']} {addr['city']}")

        # Überschriften

        c.setFont("Helvetica-Bold", 11)
        c.drawString(
            5 * mm,
            97 * mm,
            pgettext("QR-Invoice / fixed by SIX group style guide", "Empfangsschein"),
        )
        c.drawString(
            67 * mm,
            97 * mm,
            pgettext("QR-Invoice / fixed by SIX group style guide", "Zahlteil"),
        )

        # Empfangsschein / Bereich Betrag

        c.setFont("Helvetica-Bold", 6)
        c.drawString(
            5 * mm,
            33 * mm,
            pgettext("QR-Invoice / fixed by SIX group style guide", "Währung"),
        )
        c.drawString(
            20 * mm,
            33 * mm,
            pgettext("QR-Invoice / fixed by SIX group style guide", "Betrag"),
        )

        c.setFont("Helvetica", 8)
        c.drawString(5 * mm, 30 * mm, "CHF")
        if empty_total:
            _draw_corners(
                20 * mm, 32 * mm, 37 * mm, 10 * mm
            )  # Leerer Betrag (Empfangsschein); statt 30x10 hier 37x10 aus Design-Gründen
        else:
            c.drawString(20 * mm, 30 * mm, total)

        # Empfangsschein / Bereich Annahmestelle

        c.setFont("Helvetica-Bold", 6)
        c.drawRightString(
            57 * mm,
            17 * mm,
            pgettext("QR-Invoice / fixed by SIX group style guide", "Annahmestelle"),
        )

        # Empfangsschein / Bereich Angaben

        t = c.beginText(5 * mm, 90 * mm)
        _write_creditor_and_reference(t, small=True)
        _write_debitor(t, small=True)
        c.drawText(t)
        if empty_addr:
            _draw_corners(
                5 * mm,
                (59 if recv.mode == "QRR" else 68) * mm,
                52 * mm,
                20 * mm
                # this is a very hacky solution because t.getY() returns a wrong result and cannot be used
                # to determine the end of the text box
            )  # Leerer Zahlungspflichtiger

        # Zahlteil / Bereich Betrag

        c.setFont("Helvetica-Bold", 8)
        c.drawString(
            67 * mm,
            33 * mm,
            pgettext("QR-Invoice / fixed by SIX group style guide", "Währung"),
        )
        c.drawString(
            87 * mm,
            33 * mm,
            pgettext("QR-Invoice / fixed by SIX group style guide", "Betrag"),
        )

        c.setFont("Helvetica", 10)
        c.drawString(67 * mm, 29 * mm, "CHF")
        if empty_total:
            _draw_corners(
                77 * mm, 31.5 * mm, 40 * mm, 15 * mm
            )  # Leerer Betrag (Zahlteil)
        else:
            c.drawString(87 * mm, 29 * mm, total)

        # Zahlteil / Bereich Angaben

        t = c.beginText(118 * mm, 97 * mm)
        _write_creditor_and_reference(t)
        _write_additional_info(t)
        _write_debitor(t)
        c.drawText(t)
        if empty_addr:
            _draw_corners(
                118 * mm,
                (48 if recv.mode == "QRR" else 60) * mm,
                65 * mm,
                25 * mm
                # this is a very hacky solution because t.getY() returns a wrong result and cannot be used
                # to determine the end of the text box
            )  # Leerer Zahlungspflichtiger

        # Zahlteil / Bereich Weitere Informationen

        # c.setFont("Helvetica-Bold", 7)
        # c.drawString(67*mm, 11*mm, "Name AV1:")
        # c.drawString(67*mm, 8*mm, "Name AV2:")
        #
        # c.setFont("Helvetica", 7)
        # c.drawString(82*mm, 11*mm, "Linie 1")
        # c.drawString(82*mm, 8*mm, "Linie 2")

        # DEBUG

        # self.debug()

        c.restoreState()

    def draw(self):
        self.canv.translate(-12 * mm, -12 * mm)
        self.draw_qr_invoice()
